/*
	Mimicks the dependency injection system of angular.
 */
(function(){
	var fakeAngular = {};
	var registeredDependencies = {};
	var dependencies = {};

	fakeAngular.module = function(moduleName){
		//don"t bother, we only have a few dependencies to use and there shouldn't be any name collision.
		return fakeAngular;
	};

	fakeAngular.service = function(name, params){
		registeredDependencies[name] = params;
		return fakeAngular;
	};
	
	fakeAngular.getDependency = function(name){
		if(dependencies[name]){
			//already computed
			return dependencies[name];
		}else{
			var dependency = null;
			var registeredDepency = registeredDependencies[name];
			//we assume use of ng annotate
			//Note: this will probably be brittle, won't detect circular dependencies and such
			//but is ok for our modest use case
			if(typeof registeredDepency == 'function'){
				dependency = registeredDepency();
			}else if(_.isArray(registeredDepency)){
				var fun = registeredDepency.pop();
				var params = _.map(registeredDepency, function(param){
					return fakeAngular.getDependency(param);
				});
				dependency = fun.apply(null, params);
			}else{
				throw new Error('[fakeAngular] failed: getDepency('+name+')');
			}
			dependencies[name] = dependency;
			return dependency;
		}
	};
	
	window.angular = fakeAngular;
})();


//NOTE: Not an angular service, no angular dependencies, we want this to run in other contexts
//TODO: see if and how it should be packaged

(function(){

	//See: http://stackoverflow.com/questions/5775469/whats-the-valid-way-to-include-an-image-with-no-src/14115340#14115340
	var EMPTY_IMAGE_SRC = 'data:image/gif;base64,R0lGODlhAQABAAD/ACwAAAAAAQABAAACADs=';
	
	//Public API: 
	// - Player is used for preview playing in contribute tool
	// - FullSpread is used by reader-tools (for spread screenshot)
	window.fangaPreview ={
		Player: Player,
		FullSpread: FullSpread,
		EMPTY_IMAGE_SRC: EMPTY_IMAGE_SRC
	};

	
	// Call the constructor to build the player, then start() or goto() to start playing
	function Player(rootElement, model, drawUtils){
		//Model
		var spreads = model.spreads;
		var reverse = model.rightToLeft;
		var destroyed = false;
		var makeImagePath = function (image){
			return model.resourcePath + '/' + image.path;
		};

		//Layout
		var rootWidth = 0;
		var rootHeight = 0;
		var maxSpreadWidth = 0;
		var maxSpreadHeight = 0;
		//Thumbs
		var thumbs = [];
		var thumbScale = 1;
		var THUMB_MARGIN = 10;
		var THUMB_PRELOAD_MARGIN = 100;
		var THUMB_GAP = 100;
		//High res
		var highResSpreads = [];
		//Annotations are all stored in the same layer
		var annotationHolder = null;
		//Canvas overlay
		var overlayCanvas;

		//Current state (only changed via goTo)
		var zones = [];
		var currentZoneIndex = -1;
		var currentZone;
		var currentSpreadIndex = -1;
		var currentSpread;

		//Animation / Tweening
		var movingForward = true;
		var position = {};		//a can be seen as an interpolations of two zones
		var startPosition = {};
		var endPosition = {};
		var animationInProgress = false;
		var animationDuration = 300;
		var spreadAnimationDuration = 600;
		var animationEasing = function(p){
			return Math.sin(p * Math.PI / 2);
		};
		var animationStartTime;

		//Signals
		var spreadChanged = new signals.Signal();
		var zoneChanged = new signals.Signal();

		initLogger(true);

		init();

		//------------------------------------------------------------------------------------------------------------------
		// Initialization
		//------------------------------------------------------------------------------------------------------------------

		function init(){
			log('PreviewPlayer - init', model, logTime());

			rootElement.style.overflow = 'hidden';

			window.addEventListener('resize', handleResize);
			window.addEventListener('keydown', handleKeyDown);

			//define spread positions in global coordinates and create thumb objects
			var x = 0;
			for(var i = 0, len = spreads.length ; i < len ; i++){
				var spread = spreads[i];
				spread.index = i;
				//set spread position
				if(!reverse){
					spread.x = x;
					x += THUMB_GAP + spread.width;
				}else{
					spread.x = x - spread.width;
					x = spread.x - THUMB_GAP;
				}
				//keep track of max sizes
				if(spread.width > maxSpreadWidth){
					maxSpreadWidth = spread.width;
				}
				if(spread.height > maxSpreadHeight){
					maxSpreadHeight = spread.height;
				}

				//create and add thumbs
				var thumb = new SpreadThumb(spread, makeImagePath);
				rootElement.appendChild(thumb.element);
				thumbs.push(thumb);
			}

			//create high res objects
			for(i = 0; i <= 2; i++){
				var highResSpread = new FullSpread(makeImagePath, drawUtils);
				highResSpreads.push(highResSpread);
				rootElement.appendChild(highResSpread.element);
			}

			//create UI
			overlayCanvas = makeElement('canvas', null, rootElement);
			annotationHolder = makeElement('div', 'annotation-holder', rootElement);
			annotationHolder.style.transition = 'all 0.5s cubic-bezier(0.445, 0.05, 0.55, 0.95)';  //http://easings.net/#easeInOutSine
			annotationHolder.style.top = 'auto';
			annotationHolder.style.bottom = 0;
			annotationHolder.style.width = '100%';

			//precompute model shortcuts & create annotations
			for(i = 0, len = spreads.length ; i < len ; i++){
				spread = spreads[i];

				var bubbleCount = 0;
				var annotationCount = 0;
				if(!spread.panels){
					spread.panels = [];
				}
				for(var iPanel = 0; iPanel < spread.panels.length; iPanel++){
					var panel = spread.panels[iPanel];
					panel.index = iPanel;
					if(!panel.bubbles){
						panel.bubbles = [];
					}
					panel.annotations = [];
					for(var iBubble = 0; iBubble < panel.bubbles.length; iBubble++){
						var bubble = panel.bubbles[iBubble];
						bubble.index = bubbleCount;
						bubble.panel = panel;
						bubbleCount++;

						if(bubble.kind == 'point'){
							bubble.isAnnotation = true;
							bubble.annotationIndex = annotationCount;
							annotationCount++;
							panel.annotations.push(bubble);
						}else{
							bubble.isAnnotation = false;
						}
					}
				}
			}
			initWithSize();
		}
		
		// added as a convenience method, calling goto directly is fine too
		function start(spreadIndex){
			spreadIndex = spreadIndex || 0;
			goto(spreads[spreadIndex].zoneIndex, false);
		}

		function initWithSize(){
			rootWidth = rootElement.offsetWidth;
			rootHeight = rootElement.offsetHeight;
			thumbScale = Math.min(
				(rootWidth - 2 * THUMB_MARGIN) / maxSpreadWidth,
				(rootHeight - 2 * THUMB_MARGIN) / maxSpreadHeight
			);

			for(var i = 0, len = spreads.length ; i < len ; i++){
				var spread = spreads[i];
				spread.y = Math.round((maxSpreadHeight - spread.height) / 2);
			}

			overlayCanvas.width = rootWidth;
			overlayCanvas.height = rootHeight;

			computeZones();
		}

		function computeZones(){
			var margin = 20;
			zones = [];

			//create a temp div that will be used to measure annotations
			var tempAnnotationElement = makeAnnotation(rootElement);

			for(var iSpread = 0, lenSpread = spreads.length; iSpread < lenSpread; iSpread++){
				var spread = spreads[iSpread];
				var spreadZone = {
					spreadIndex: iSpread,
					index: zones.length,
					isFull: true,
					rectX: spread.x,
					rectY: spread.y,
					rectWidth: spread.width,
					rectHeight: spread.height,
					annotationHeight: 0,
					scale: thumbScale
				};
				spread.zoneIndex = spreadZone.index;
				zones.push(spreadZone);

				var zone = null;
				for(var iPanel = 0, lenPanel = spread.panels.length; iPanel < lenPanel; iPanel++){
					var panel = spread.panels[iPanel];
					var annotationHeight = 0;
					for(var iAnnotation = 0; iAnnotation < panel.annotations.length; iAnnotation++){
						var annotationBubble = panel.annotations[iAnnotation];
						applyAnnotation(tempAnnotationElement, annotationBubble);
						annotationHeight +=  tempAnnotationElement.offsetHeight;
					}

					//we aready have a zone, can we merge with it
					var panelX = spread.x + panel.x * spread.width;
					var panelY = spread.y + panel.y * spread.height;
					var panelWidth = panel.width * spread.width;
					var panelHeight = panel.height * spread.height;
					if(zone){
						var panelScale = Math.min(
							1,
							(rootWidth - 2 * margin) / panelWidth ,
							(rootHeight - 2 * margin - annotationHeight) / panelHeight
						);
						var mergeX = Math.min(zone.rectX, panelX);
						var mergeY = Math.min(zone.rectY, panelY);
						var mergeWidth = Math.max(zone.rectX + zone.rectWidth, panelX + panelWidth) - mergeX;
						var mergeHeight = Math.max(zone.rectY + zone.rectHeight, panelY + panelHeight) - mergeY;
						var mergeScale = Math.min(
							1,
							(rootWidth - 2 * margin) / mergeWidth ,
							(rootHeight - 2 * margin - zone.annotationHeight - annotationHeight) / mergeHeight
						);
						var canMerge = false;
						if(mergeScale == 1){
							//yes, we can merge and still display at 1:1
							canMerge = true;
						}else if(Math.abs(zone.scale - mergeScale) < 0.1 && Math.abs(panelScale - mergeScale) < 0.1){
							//we can not be displayed at 1:1, and merging doesn't change effective scale that much
							canMerge = true;
						}
						if(canMerge){
							zone.panels.push(panel);
							zone.rectX = mergeX;
							zone.rectY = mergeY;
							zone.rectWidth = mergeWidth;
							zone.rectHeight = mergeHeight;
							zone.annotationHeight += annotationHeight;
							zone.annotations = zone.annotations.concat(panel.annotations);
						}else{
							zone = null;
						}
					}
					if(!zone){
						zone = {
							spreadIndex: iSpread,
							index: zones.length,
							panels: [panel],
							rectX: panelX,
							rectY: panelY,
							rectWidth: panelWidth,
							rectHeight: panelHeight,
							annotationHeight: annotationHeight,
							annotations: panel.annotations
						};
						zone.scale = Math.min(
							1,
							(rootWidth - 2 * margin) / zone.rectWidth ,
							(rootHeight - 2 * margin - zone.annotationHeight) / zone.rectHeight
						);

						zones.push(zone);
					}
				}
			}

			//Compute actual positions so that the zones are centered in screen
			for(var iZone = 0, lenZone = zones.length; iZone < lenZone; iZone++){
				zone = zones[iZone];
				spread = spreads[zone.spreadIndex];
				//compute screen position in global coordinates
				var sw = rootWidth / zone.scale;
				var sh = (rootHeight - zone.annotationHeight) / zone.scale;
				var sx = (zone.rectX + zone.rectWidth / 2) - sw / 2;
				var sy = (zone.rectY + zone.rectHeight / 2) - sh / 2;

				//keep screen rect inside spread
				if(sx < spread.x - margin && sx + sw < spread.x + spread.width){
					//overflow left
					sx = spread.x - margin;
				}else if(sx + sw > spread.x + spread.width + margin && sx > spread.x){
					//overflow right
					sx = spread.x + spread.width + margin - sw;
				}

				if(sy < spread.y - margin && sy + sh < spread.y + spread.height){
					//overflow top
					sy = spread.y - margin;
				}else if(sy + sh > spread.y + spread.height + margin && sy > spread.y){
					//overflow bottom
					sy = spread.y + spread.height + margin - sh;
				}
				zone.x = sx;
				zone.y = sy;
			}

			rootElement.removeChild(tempAnnotationElement);
		}

		function handleResize(){
			if(handleResize.timeoutId){
				clearTimeout(handleResize.timeoutId);
			}
			handleResize.timeoutId = setTimeout(function(){
				initWithSize();

				if(currentZone){
					//try to show the same place
					if(!currentZone.isFull){
						//show the zone that contains the same panel
						var panel = currentZone.panels[0];
						for(var i = 0; i < zones.length; i++){
							var zone = zones[i];
							if(zone.spreadIndex == currentZone.spreadIndex &&
								!zone.isFull &&
								zone.panels.indexOf(currentZone.panels[0]) >= 0){
								//zone found
								goto(i);
								return;
							}
						}
					}
					//show same spread
					goto(spreads[currentZone.spreadIndex].zoneIndex);
				}else{
					//fall back to the beginning
					goto(0);
				}
			}, 100);
		}

		//------------------------------------------------------------------------------------------------------------------
		// Animation & display
		//------------------------------------------------------------------------------------------------------------------

		function goto(zoneIndex, doAnim){
			log('PreviewPlayer - goto', zoneIndex, !!doAnim, logTime());

			var prevZone = currentZone;
			var prevSpreadIndex = currentSpreadIndex;

			currentZoneIndex = zoneIndex;
			currentZone = zones[currentZoneIndex];

			updateAnnotations();

			if(currentSpreadIndex != currentZone.spreadIndex){
				currentSpreadIndex = currentZone.spreadIndex;
				currentSpread = spreads[currentSpreadIndex];
				updateHighResLater();
			}

			if(!doAnim){
				position.zone = currentZone;
				position.x = currentZone.x;
				position.y = currentZone.y;
				position.scale = currentZone.scale;
				updateDisplay();
			}else{
				startPosition.zone = position.zone;
				startPosition.x = position.x;
				startPosition.y = position.y;
				startPosition.scale = position.scale;
				endPosition.zone = currentZone;
				endPosition.x = currentZone.x;
				endPosition.y = currentZone.y;
				endPosition.scale = currentZone.scale;

				animationStartTime = Date.now();
				requestTic();
			}

			//signals
			if(currentSpreadIndex != prevSpreadIndex){
				spreadChanged.dispatch(currentSpread);
			}
			if(currentZone != prevZone){
				zoneChanged.dispatch(currentZone);
			}
		}

		function requestTic(){
			//console.log('requestAnimationFrame');
			window.requestAnimationFrame(tic, rootElement);
		}

		//computes intermediary positions during the animation
		function tic(){
			if(destroyed){
				return
			}

			if(position.scale != endPosition.scale || position.x != endPosition.x || position.y != endPosition.y){
				animationInProgress = true;
				position.zone = endPosition.zone;

				var duration;
				if(startPosition.zone.spreadIndex != endPosition.zone.spreadIndex){
					duration = spreadAnimationDuration;
				}else{
					duration = animationDuration;
				}
				var p = (Date.now() - animationStartTime) / duration;
				if(p < 0){
					p = 0;
				}
				if(p >= 1){
					p = 1;
					animationInProgress = false;
					position.scale = endPosition.scale;
					position.x = endPosition.x;
					position.y = endPosition.y;
				}else{
					p = animationEasing(p);
					position.scale = (1 - p) * startPosition.scale + p * endPosition.scale;

					// "naive" interpolation looks weird because it doesn't consider camera/screen position
					//position.x = (1-p) * startPosition.x + p * endPosition.x;
					//position.y = (1-p) * startPosition.y + p * endPosition.y;

					// we can't interpolate global position directly, we interpolate the spread position relative
					// to the screen instead
					var startSpreadX = (currentSpread.x - startPosition.x) * startPosition.scale;
					var startSpreadY = (currentSpread.y - startPosition.y) * startPosition.scale;
					var endSpreadX = (currentSpread.x - endPosition.x) * endPosition.scale;
					var endSpreadY = (currentSpread.y - endPosition.y) * endPosition.scale;
					// interpolate
					var spreadX = (1 - p) * startSpreadX + p * endSpreadX;
					var spreadY = (1 - p) * startSpreadY + p * endSpreadY;

					position.x = currentSpread.x - spreadX / position.scale;
					position.y = currentSpread.y - spreadY / position.scale;
				}
			}else{
				animationInProgress = false;
			}

			updateDisplay();

			if(animationInProgress){
				//TODO only tic while animating
				requestTic();
			}

		}

		//layouts the DOM and redraws the overlay canvas
		function updateDisplay(){
			// Update thumbs
			for(var i = 0, len = thumbs.length; i < len; i++){
				var spread = spreads[i];

				var x = Math.round((spread.x - position.x) * position.scale);
				var y = Math.round((spread.y - position.y) * position.scale);
				var width = spread.width * position.scale;

				var thumb = thumbs[i];
				if(x + width < -THUMB_PRELOAD_MARGIN || x > rootWidth + THUMB_PRELOAD_MARGIN){
					//out of range or hidden by high res
					if(thumb.visible){
						thumb.hide();
					}
				}else{
					if(!thumb.visible){
						thumb.show();
					}
					var scaleX = position.scale;
					var scaleY = position.scale;
					if(!animationInProgress){
						//make sure we get round pixel values
						x = Math.round(x);
						y = Math.round(y);
						//scaleX = Math.round(spread.width * scaleX) / spread.width;
						//scaleY = Math.round(spread.height * scaleY) / spread.height;
					}
					thumb.x = x;
					thumb.y = y;
					thumb.scaleX = scaleX;
					thumb.scaleY = scaleY;

					setPrefixedProp(
						thumb.element,
						'transform',
						'translate3d(' + x + 'px, ' + y + 'px, 0px) scale(' + scaleX + ',' + scaleY + ')');
				}
			}

			//Update High Res Spreads
			for(i = 0, len = highResSpreads.length; i < len; i++){
				var highResSpread = highResSpreads[i];
				spread = highResSpread.spread;
				if(spread){
					thumb = thumbs[spread.index];
					highResSpread.setIsVisibleInScreen(thumb.visible);
					if(thumb.visible){
						var transform = 'translate3d(' + thumb.x + 'px, ' + thumb.y + 'px, 0px) scale(' + thumb.scaleX + ',' + thumb.scaleY + ')';
						setPrefixedProp(highResSpread.element, 'transform', transform);

						if(spread.index == currentSpreadIndex){
							highResSpread.setZone(currentZone);
						}else{
							highResSpread.setZone(null);
						}
					}
				}
			}

			//Update canvas
			var ctx = overlayCanvas.getContext('2d');
			ctx.save();
			ctx.clearRect(0, 0, rootWidth, rootHeight);

			ctx.fillStyle = 'rgba(32, 159, 203, 0.5)';
			ctx.fillRect(0, 0, rootWidth, rootHeight);

			if(!currentZone.panels){
				ctx.clearRect(
					(currentZone.rectX - position.x) * position.scale,
					(currentZone.rectY - position.y) * position.scale,
					(currentZone.rectWidth) * position.scale,
					(currentZone.rectHeight) * position.scale
				);
			}else{
				ctx.strokeStyle = '#ffc036';
				ctx.lineWidth = 2;
				ctx.beginPath();
				for(var iPanel = 0, lenPanel = currentZone.panels.length; iPanel < lenPanel; iPanel++){
					var panel = currentZone.panels[iPanel];
					thumb = thumbs[currentSpreadIndex];

					var panelX = currentSpread.x + panel.x * currentSpread.width - position.x;
					var panelY = currentSpread.y + panel.y * currentSpread.height - position.y;
					var panelWidth = panel.width * currentSpread.width;
					var panelHeight = panel.height * currentSpread.height;
					panelX *= position.scale;
					panelY *= position.scale;
					panelWidth *= position.scale;
					panelHeight *= position.scale;
					ctx.clearRect(panelX, panelY, panelWidth, panelHeight);
					ctx.rect(panelX, panelY, panelWidth, panelHeight);
				}
				ctx.stroke();
			}
			ctx.restore();
		}

		function updateHighResLater(){
			if(updateHighResLater.timeoutId){
				clearTimeout(updateHighResLater.timeoutId);
			}
			updateHighResLater.timeoutId = setTimeout(updateHighRes, 500);
		}

		function updateHighRes(){
			var spreadsToLoad = [
				spreads[currentSpreadIndex - 1],
				spreads[currentSpreadIndex],
				spreads[currentSpreadIndex + 1]
			];
			//find which spreads are already displayed
			var remainingHighResSpreads = [];
			for(var i = 0 ; i < highResSpreads.length; i++){
				var highResSpread = highResSpreads[i];
				var keep = false;
				for(var j = 0; j < spreadsToLoad.length; j++){
					if(spreadsToLoad[j] && spreadsToLoad[j] == highResSpread.spread){
						keep = true;
						break;
					}
				}
				if(keep){
					spreadsToLoad[j] = null;
				}else{
					remainingHighResSpreads.push(highResSpread);
				}
			}
			//give other spreads to loads to remaining highResSpreads
			for(i = 0 ; i < remainingHighResSpreads.length; i++){
				var found = false;
				highResSpread = remainingHighResSpreads[i];
				for(j = 0; j < spreadsToLoad.length; j++){
					var spread = spreadsToLoad[j];
					if(spread){
						found = true;
						highResSpread.setSpread(spread);
						spreadsToLoad[j] = null;
						break;
					}
				}
				if(!found){
					highResSpread.setSpread(null);
				}
			}
			updateDisplay();
		}

		//------------------------------------------------------------------------------------------------------------------
		// Annotations
		//------------------------------------------------------------------------------------------------------------------

		function updateAnnotations(){
			var numAnnotations = 0;
			if(currentZone && currentZone.annotations){
				numAnnotations = currentZone.annotations.length;
			}

			if(numAnnotations == 0){
				//disappear animation if already displayed
				annotationHolder.style.opacity = 0;
				annotationHolder.style.bottom = '-100px';
			}else{
				//triggers appear animation if not displayed yet
				annotationHolder.style.opacity = 1;
				annotationHolder.style.bottom = 0;

				//hide unnecessary elements
				var children = annotationHolder.childNodes;
				for(var i = numAnnotations; i < children.length; i++){
					children[i].style.display = 'none';
				}
				//create missing elements
				for(i = children.length; i < numAnnotations; i++){
					makeAnnotation(annotationHolder);
				}
				//update displayed elements
				children = annotationHolder.childNodes;
				for(i = 0; i < numAnnotations; i++){
					var bubble = currentZone.annotations[i];
					var annotationElement = children[i];
					var style = annotationElement.style;
					style.display = 'block';
					applyAnnotation(annotationElement, bubble, i);
				}
			}
		}

		//------------------------------------------------------------------------------------------------------------------
		// Navigation
		//------------------------------------------------------------------------------------------------------------------

		function goToNextSpread(){
			movingForward = true;
			if(currentSpreadIndex < spreads.length - 1){
				goto(spreads[currentSpreadIndex + 1].zoneIndex, true);
			}
		}

		function goToPrevSpread(){
			movingForward = false;
			if(currentSpreadIndex > 0){
				goto(spreads[currentSpreadIndex - 1].zoneIndex, true);
			}
		}

		function goToSpread(index, animate){
			if(index < 0){
				index = 0;
			}else if(index >= spreads.length){
				index = spreads.length - 1;
			}
			goto(spreads[index].zoneIndex, animate);
		}

		function goToNextZone(){
			movingForward = true;
			if(currentZoneIndex < zones.length - 1){
				goto(currentZoneIndex + 1, true);
			}
		}

		function goToPrevZone(){
			movingForward = false;
			if(currentZoneIndex > 0){
				goto(currentZoneIndex - 1, true);
			}
		}

		function hasNextZone(){
			return currentZone && zones && currentZone.index < zones.length - 1;
		}

		function hasPrevZone(){
			return currentZone && zones && currentZone.index > 0;
		}

		function handleKeyDown(e){
			if(e.keyCode == '38'){
				// up arrow
				goToPrevSpread();
			}
			else if(e.keyCode == '40'){
				// down arrow
				goToNextSpread();
			}
			else if(e.keyCode == '37'){
				// left arrow
				goLeft();
			}
			else if(e.keyCode == '39'){
				goRight();
			}
		}

		function goLeft(){
			if(!reverse){
				goToPrevZone();
			}else{
				goToNextZone();
			}
		}

		function goRight(){
			if(!reverse){
				goToNextZone();
			}else{
				goToPrevZone();
			}
		}

		function canGoLeft(){
			if(!reverse){
				return hasPrevZone();
			}else{
				return hasNextZone();
			}
		}

		function canGoRight(){
			if(!reverse){
				return hasNextZone();
			}else{
				return hasPrevZone();
			}
		}

		//------------------------------------------------------------------------------------------------------------------
		// Public Player API
		//------------------------------------------------------------------------------------------------------------------
		var self = {};
		
		self.play = function(){
			console.log('Player.play');
		};

		self.pause = function(){
			console.log('Player.pause');
		};

		self.restart = function(){
			console.log('Player.restart');
		};

		self.destroy = function(){
			destroyed = true;

			while(rootElement.firstChild){
				rootElement.removeChild(rootElement.firstChild);
			}

			window.removeEventListener('resize', handleResize);
			window.removeEventListener('keydown', handleKeyDown);
		};

		//nav
		self.start = start;
		self.goToNextZone = goToNextZone;
		self.goToPrevZone = goToPrevZone;
		self.goLeft = goLeft;
		self.goRight = goRight;
		self.canGoLeft = canGoLeft;
		self.canGoRight = canGoRight;
		self.goToSpread = goToSpread;

		//signals
		self.spreadChanged = spreadChanged;
		self.zoneChanged = zoneChanged;

		//For debug, gives console access
		Player.instance = self;
		
		return self;
	}

	//------------------------------------------------------------------------------------------------------------------
	// Thumb spread
	//------------------------------------------------------------------------------------------------------------------

	function SpreadThumb(spread, makeImagePath){
		this.makeImagePath = makeImagePath;
		this.spread = spread;

		this.element = makeElement('div', 'spread-thumb');
		setPrefixedProp(this.element, 'transformOrigin', '0 0');
		this.element.style.background = 'white';
		this.element.style.visibility = 'hidden';
		this.element.setAttribute('spread', spread.index);
		this.element.style.width = spread.width + 'px';
		this.element.style.height = spread.height + 'px';

		this.imgElements = [
			makeElement('img', 'thumb', this.element),
			makeElement('img', 'thumb', this.element)
		];
		for(var i = 0; i < this.spread.pages.length ; i++){
			var page = this.spread.pages[i];
			var imgElement = this.imgElements[i];
			var style = imgElement.style;
			//place
			style.left = (page.x * 100) + '%';
			style.top = (page.y * 100) + '%';
			style.width = (page.width * 100) + '%';
			style.height = (page.height * 100) + '%';
			style.opacity = 0;
		}

		this.visible = false;
		this.x = 0;
		this.y = 0;
		this.width = spread.width;
		this.height = spread.height;

		this.loaded = false;
		this.loading = false;
	}

	SpreadThumb.prototype.show = function(){
		if(!this.visible){
			if(!this.loading && !this.loaded){
				this._loadLater();
			}
			this.visible = true;
			this.element.style.visibility = 'visible';
		}
	};

	SpreadThumb.prototype.hide = function(){
		if(this.visible){
			this.visible = false;
			this.element.style.visibility = 'hidden';
		}
	};

	SpreadThumb.prototype._loadLater = function(){
		var self = this;

		if(this._loadlaterTimeoutId){
			return;
		}
		this._loadlaterTimeoutId = setTimeout(function(){
			self._loadlaterTimeoutId = null;
			if(self.visible){
				self._load();
			}
		}, 500);
	};

	SpreadThumb.prototype._load = function(){
		if(this.loading || this.loaded){
			return;
		}
		//console.log('FangaSpreadThumb._load',this.spread.index);
		this.loading = true;
		this.imageLoadedCount = 0;

		for(var i = 0; i < this.spread.pages.length; i++){
			var page = this.spread.pages[i];
			var imgElement = this.imgElements[i];
			setPrefixedProp(imgElement, 'transition', 'opacity 0.5s');

			imgElement.onload = makeCallback(this._onImageLoaded, this, [i]);
			imgElement.src = this.makeImagePath(page.thumbnail);
		}
	};

	SpreadThumb.prototype._onImageLoaded = function(pageIndex){
		var imgElement = this.imgElements[pageIndex];
		imgElement.style.opacity = 1;

		this.imageLoadedCount++;
		if(this.imageLoadedCount == this.spread.pages.length){
			this.loading = false;
			this.loaded = true;
		}
	};

	//------------------------------------------------------------------------------------------------------------------
	// Full Spread
	//------------------------------------------------------------------------------------------------------------------

	FullSpread.idCount = 0;
	function FullSpread(makeImagePath, drawUtils){
		this.makeImagePath = makeImagePath;
		this.drawUtils = drawUtils;
		this.id = 'FullSpread' + FullSpread.idCount;
		FullSpread.idCount++;

		this.element = makeElement('div', 'full-spread');
		this.element.style.overflow = 'hidden';
		this.element.style.visibility = 'hidden';
		setPrefixedProp(this.element, 'transformOrigin', '0 0');

		this.imgElements = [
			makeElement('img', 'full-image', this.element),
			makeElement('img', 'full-image', this.element)
		];
		this.canvas = makeElementWithId('canvas', 'full-canvas', 'box', this.element);
		this.spreadCanvas = makeElement('canvas');

		this.x = 0;
		this.y = 0;
		this.spread = null;
		this.zone = null;
		this.annotationCircles = [];	//div elements, they are cached for reuse, see setZone
		this.isVisibleInScreen = false;
		this.spreadDisplayed = new signals.Signal();

		this._updateVisibility();
	}

	FullSpread.prototype.setIsVisibleInScreen = function(visible){
		if(visible != this.isVisibleInScreen){
			this.isVisibleInScreen = visible;
			this._updateVisibility();
		}
	};

	FullSpread.prototype.setSpread = function(spread){
//		if(spread == this.spread){
//			return;
//		}
		this.spread = spread;
		this.setZone(null);

		this.element.setAttribute('spread', spread ? spread.index : 'none');
		this.isVisibleInScreen = false;
		this.loaded = false;
		this._updateVisibility();

		if(spread){
			this.imageLoadedCount = 0;
			this.width = spread.width;
			this.height = spread.height;
			this.element.style.width = spread.width + 'px';
			this.element.style.height = spread.height + 'px';

			for(var i = 0; i < this.imgElements.length; i++){
				var imgElement = this.imgElements[i];
				var page = this.spread.pages[i];
				var style = imgElement.style;
				if(page){
					var src = this.makeImagePath(page.image);
					if(imgElement._src == src && imgElement.complete){
						this._onImageLoaded(i, spread);
					}else{
						imgElement.onload = makeCallback(this._onImageLoaded, this, [i, this.spread]);
						imgElement.src = src;
						imgElement._src = src;
					}
					style.position = 'absolute';
					style.left = (page.x * this.spread.width) + 'px';
					style.top = (page.y * this.spread.height) + 'px';
					//NOTE: for the moment we don't specify width & height since pages are never scaled, 
					// but the model makes it possible in theory, so we may need to change that later on.
					
					style.display = 'block';
				}else{
					style.display = 'none';
				}
			}
			if(this.spread.pages.length === 0){
				this._onAllImageLoaded();
			}
		}
	};

	FullSpread.prototype._onImageLoaded = function(pageIndex, spread){
		if(spread == this.spread){
			this.imageLoadedCount++;
			if(this.imageLoadedCount == this.spread.pages.length){
				this._onAllImageLoaded();
			}
		}
	};

	FullSpread.prototype._onAllImageLoaded = function(){
		this.loading = false;
		this.loaded = true;
		this._drawCanvas();
		this._displayTexts();
		this._updateVisibility();

		this.spreadDisplayed.dispatch(this.spread);
	};

	FullSpread.prototype._drawCanvas = function(){
		var ctx;
		//Draw the page images in a separate canvas
		ctx = this.spreadCanvas.getContext('2d');
		this.spreadCanvas.width = this.spread.width;
		this.spreadCanvas.height = this.spread.height;

		for(var i = 0, len = this.spread.pages.length; i < len; i++){
			var img = this.imgElements[i];
			var page = this.spread.pages[i];
			ctx.drawImage(img,
				0, 0, img.naturalWidth, img.naturalHeight,
				page.x * this.spread.width,
				page.y * this.spread.height,
				page.width * this.spread.width,
				page.height * this.spread.height
			);
		}

		//Draw only the bubbles in another canvas
		//Ideally, we'd draw everything in a single canvas but it scales down poorly.
		//Instead we display the full page with img tags and a big canvas with only the bubbles on top of it.
		ctx = this.canvas.getContext('2d');
		this.canvas.width = this.spread.width;
		this.canvas.height = this.spread.height;
		for(i = 0, len = this.spread.panels.length; i < len; i++){
			var panel = this.spread.panels[i];
			for(var j = 0, lenj = panel.bubbles.length; j < lenj; j++){
				var bubble = panel.bubbles[j];
				if(bubble.kind != 'point'){
					this.drawUtils.drawBubbleBackground(bubble, panel, this.spread, {
						x: (panel.x + bubble.x * panel.width) * this.spread.width,
						y: (panel.y + bubble.y * panel.height) * this.spread.height,
						width: (bubble.width * panel.width) * this.spread.width,
						height: (bubble.height * panel.height) * this.spread.height
					}, this.spreadCanvas, ctx);
				}
			}
		}
	};

	FullSpread.prototype._updateVisibility = function(){
		if(this.isVisibleInScreen && this.loaded){
			this.element.style.visibility = 'visible';
		}else{
			this.element.style.visibility = 'hidden';
		}
	};

	FullSpread.prototype._displayTexts = function(){
		if(!this.textDiv){
			this.textDiv = makeElement('div', 'text-div', this.element);
			this.textDiv.style.width = '100%';
			this.textDiv.style.height = '100%';

			//fixes texts not being rendered after going out of screen and coming back in Chrome
			setPrefixedProp(this.textDiv, 'transform', 'translateZ(0)');
		}else{
			this.textDiv.innerHTML = '';
		}

		var fragment = document.createDocumentFragment();
		for(var i = 0, len = this.spread.panels.length; i < len; i++){
			var panel = this.spread.panels[i];
			for(var j = 0, lenj = panel.bubbles.length; j < lenj; j++){
				var bubble = panel.bubbles[j];
				if(!bubble.isAnnotation && bubble.translation && bubble.translation.frame && bubble.translation.text && bubble.translation.textStyle){
					//Create frame
					var frame = bubble.translation.frame;
					var frameElement = makeElement('div', null, fragment);
					var style = frameElement.style;
					style.display = 'table';
					style.left = frame.x * this.spread.width + 'px';
					style.top = frame.y * this.spread.height + 'px';
					style.width = frame.width * this.spread.width + 'px';
					style.height = frame.height * this.spread.height + 'px';

					var textElement = makeElement('div', null, frameElement);
					style = textElement.style;
					style.position = 'initial';
					style.display = 'table-cell';
					textElement.innerHTML = bubble.translation.text;
					style.color = bubble.bubbleStyle.textColor;
					style.verticalAlign = bubble.translation.textStyle.verticalAlign;
					style.textAlign = bubble.translation.textStyle.horizontalAlign;
					style.fontSize = bubble.translation.textStyle.fontSize + 'px';
					style.fontFamily = bubble.translation.textStyle.fontFamily;
				}
			}
		}
		this.textDiv.appendChild(fragment);
	};

	//When zone changes, we update the annotation elements position
	FullSpread.prototype.setZone = function(zone){
		if(zone != this.zone){
			this.zone = zone;

			var numAnnotations = 0;
			if(zone && zone.annotations){
				numAnnotations = zone.annotations.length;
			}

			//hide unnecessary elements
			for(var i = numAnnotations; i < this.annotationCircles.length; i++){
				this.annotationCircles[i].style.opacity = 0;
			}
			//create missing elements
			for(i = this.annotationCircles.length; i < numAnnotations; i++){
				var annotationCircle = makeElement('div', 'annotation-circle', this.element);
				setAnnotationCircleStyle(annotationCircle);
				//animate opacity
				annotationCircle.style.opacity = 0;
				annotationCircle.style.transition = 'opacity 0.5s cubic-bezier(0.445, 0.05, 0.55, 0.95)';  //http://easings.net/#easeInOutSine

				this.annotationCircles.push(annotationCircle);
			}
			//update displayed elements
			for(i = 0; i < numAnnotations; i++){
				var bubble = zone.annotations[i];
				annotationCircle = this.annotationCircles[i];
				var style = annotationCircle.style;
				style.opacity = 1;
				style.left = ((bubble.panel.x + bubble.x * bubble.panel.width) * this.spread.width) + 'px';
				style.top =  ((bubble.panel.y + bubble.y * bubble.panel.height) * this.spread.height) + 'px';
				annotationCircle.innerHTML = (i + 1) + '';
			}
		}
	};

	//------------------------------------------------------------------------------------------------------------------
	// Common for annotations
	//------------------------------------------------------------------------------------------------------------------

	function makeAnnotation(parent){
		var annotationElement = makeElement('div', 'annotation', parent);

		var annotationText = makeElement('div', 'annotation-text', annotationElement);
		//leave room for the circle
		annotationText.style.paddingTop = '4px';
		annotationText.style.paddingLeft = '34px';
		annotationText.style.paddingBottom = '4px';
		annotationText.style.minHeight = '32px';
		annotationText.style.position = 'relative';
		annotationElement['_fanga_annotationText'] = annotationText;

		var annotationCircle = makeElement('div', 'annotation-circle', annotationElement);
		setAnnotationCircleStyle(annotationCircle);
		annotationCircle.style.margin = '2px';
		annotationElement['_fanga_annotationCircle'] = annotationCircle;

		return annotationElement;
	}

	function applyAnnotation(annotationElement, bubble, index){
		var style = annotationElement.style;
		style.position = 'relative';
		style.background = 'rgba(255,255,255,0.8)';
		style.color = 'black';
		style.margin = '2px 5px';
		if(bubble.translation && bubble.translation.textStyle){
			style.fontSize = bubble.translation.textStyle.fontSize + 'px';
			style.fontFamily = bubble.translation.textStyle.fontFamily;
		}

		index = index || 0;
		if(bubble.translation && bubble.translation.text){
			annotationElement['_fanga_annotationText'].innerHTML = bubble.translation.text;
		}
		annotationElement['_fanga_annotationCircle'].innerHTML = (index + 1) + '';
	}

	function setAnnotationCircleStyle(annotationElement){
		var style = annotationElement.style;
		var radius = 14;
		style.position = 'absolute';
		style.margin = -radius + 'px';

		style.width = (2 * radius) + 'px';
		style.height = (2 * radius) + 'px';
		style.borderRadius = radius + 'px';

		style.border = '2px solid white';
		style.color = 'white';
		style.backgroundColor = 'rgb(32,159,203)';

		style.textAlign = 'center';
		style.fontSize = '14px';
		style.lineHeight = (1.8 * radius) + 'px';
		style.fontFamily = 'Arial, sans';
	}

	//------------------------------------------------------------------------------------------------------------------
	// Common Utils
	//------------------------------------------------------------------------------------------------------------------

	

	function makeCallback(fun, self, args){
		return function(){
			fun.apply(self, args);
		}
	}

	function makeElement(type, className, parent){
		var element = document.createElement(type);
		element.style.position = 'absolute';
		element.style.left = '0';
		element.style.top = '0';
		if(className){
			element.className = className;
		}
		if(parent){
			parent.appendChild(element);
		}
		return element;
	}

	function makeElementWithId(type, className, id, parent){
		var element = document.createElement(type);
		element.style.position = 'absolute';
		element.style.left = '0';
		element.style.top = '0';
		if(className){
			element.className = className;
		}
		if(id){
            element.id = id;
        }
		if(parent){
			parent.appendChild(element);
		}
		return element;
	}

	//cache prefixed CSS props
	function setPrefixedProp(element, prop, value){
		if(setPrefixedProp.cache[prop]){
			prop = setPrefixedProp.cache[prop];
		}else{
			prop = setPrefixedProp.findPrefixedCssProp(prop);
		}
		element.style[prop] = value;
	}

	// hide implementation details of setPrefixedProp
	(function(){
		var element = document.createElement('div');
		var prefixes = ['Webkit', 'Moz', 'ms', 'O'];
		setPrefixedProp.cache = {};
		//eg: pass in: 'transformOrigin'
		//=>  looks for 'transformOrigin', 'WebkitTransformOrigin', 'MozTransformOrigin',.. in style declaration
		setPrefixedProp.findPrefixedCssProp = function(cssProp){
			var result = cssProp;
			//if cssProp doesn't exist, try to find its prefixed version
			if(!cssProp in element.style){
				//capitalize first letter
				var capitalizedCssProp = cssProp.charAt(0).toUpperCase() + cssProp.slice(1);
				for(var i = 0; i < prefixes.length; i++){
					var prefixedProp = prefixes[i] + capitalizedCssProp;
					if(prefixedProp in element.style){
						result = prefixedProp;
						break;
					}
				}
			}
			setPrefixedProp.cache[cssProp] = result;
			return result;
		}
	})();

	//------------------------------------------------------------------------------------------------------------------
	// Log & Debug
	//------------------------------------------------------------------------------------------------------------------

	var log;
	var logTime;
	function initLogger(doLog){
		if(!initLogger._loggerLogWrapper){
			initLogger._loggerLogWrapper = Function.prototype.bind.call(console.log, console);
		}
		var loggerStartTime = Date.now();
		if(doLog){
			log = initLogger._loggerLogWrapper;
			logTime = function(){ return '[' + (Date.now() - loggerStartTime) + 'ms]'; };
		}else{
			log = logTime = function(){};
		}
	}

	//------------------------------------------------------------------------------------------------------------------
	// Signals
	//------------------------------------------------------------------------------------------------------------------

	// This is now also part of the Fangaka build
	// TODO: define preview player dependencies and have a separate build task ?

	//jscs:disable
	/*
	 JS Signals <http://millermedeiros.github.com/js-signals/>
	 Released under the MIT license
	 Author: Miller Medeiros
	 Version: 1.0.0 - Build: 268 (2012/11/29 05:48 PM)
	 */
	(function(i){function h(a,b,c,d,e){this._listener=b;this._isOnce=c;this.context=d;this._signal=a;this._priority=e||0}function g(a,b){if(typeof a!=="function")throw Error("listener is a required param of {fn}() and should be a Function.".replace("{fn}",b));}function e(){this._bindings=[];this._prevParams=null;var a=this;this.dispatch=function(){e.prototype.dispatch.apply(a,arguments)}}h.prototype={active:!0,params:null,execute:function(a){var b;this.active&&this._listener&&(a=this.params?this.params.concat(a):
		a,b=this._listener.apply(this.context,a),this._isOnce&&this.detach());return b},detach:function(){return this.isBound()?this._signal.remove(this._listener,this.context):null},isBound:function(){return!!this._signal&&!!this._listener},isOnce:function(){return this._isOnce},getListener:function(){return this._listener},getSignal:function(){return this._signal},_destroy:function(){delete this._signal;delete this._listener;delete this.context},toString:function(){return"[SignalBinding isOnce:"+this._isOnce+
		", isBound:"+this.isBound()+", active:"+this.active+"]"}};e.prototype={VERSION:"1.0.0",memorize:!1,_shouldPropagate:!0,active:!0,_registerListener:function(a,b,c,d){var e=this._indexOfListener(a,c);if(e!==-1){if(a=this._bindings[e],a.isOnce()!==b)throw Error("You cannot add"+(b?"":"Once")+"() then add"+(!b?"":"Once")+"() the same listener without removing the relationship first.");}else a=new h(this,a,b,c,d),this._addBinding(a);this.memorize&&this._prevParams&&a.execute(this._prevParams);return a},
		_addBinding:function(a){var b=this._bindings.length;do--b;while(this._bindings[b]&&a._priority<=this._bindings[b]._priority);this._bindings.splice(b+1,0,a)},_indexOfListener:function(a,b){for(var c=this._bindings.length,d;c--;)if(d=this._bindings[c],d._listener===a&&d.context===b)return c;return-1},has:function(a,b){return this._indexOfListener(a,b)!==-1},add:function(a,b,c){g(a,"add");return this._registerListener(a,!1,b,c)},addOnce:function(a,b,c){g(a,"addOnce");return this._registerListener(a,
			!0,b,c)},remove:function(a,b){g(a,"remove");var c=this._indexOfListener(a,b);c!==-1&&(this._bindings[c]._destroy(),this._bindings.splice(c,1));return a},removeAll:function(){for(var a=this._bindings.length;a--;)this._bindings[a]._destroy();this._bindings.length=0},getNumListeners:function(){return this._bindings.length},halt:function(){this._shouldPropagate=!1},dispatch:function(a){if(this.active){var b=Array.prototype.slice.call(arguments),c=this._bindings.length,d;if(this.memorize)this._prevParams=
			b;if(c){d=this._bindings.slice();this._shouldPropagate=!0;do c--;while(d[c]&&this._shouldPropagate&&d[c].execute(b)!==!1)}}},forget:function(){this._prevParams=null},dispose:function(){this.removeAll();delete this._bindings;delete this._prevParams},toString:function(){return"[Signal active:"+this.active+" numListeners:"+this.getNumListeners()+"]"}};var f=e;f.Signal=e;typeof define==="function"&&define.amd?define(function(){return f}):typeof module!=="undefined"&&module.exports?module.exports=f:i.signals=
		f})(this);

})();
//jscs:disable
(function(){
angular.module('common.utils').service('canvasBlurUtils',
	function(){
		return {
			blur: stackBlurImageFn
			//blur: stackBoxBlurImageFn
			//blur: boxBlurImageFn
		};
	}
);

//TODO: comment out the algorithms we don't actually use
var boxBlurImageFn;
var stackBlurImageFn;
var stackBoxBlurImageFn;
	

(function(){

//http://www.quasimondo.com/StackBlurForCanvas/StackBlur.js
//http://www.quasimondo.com/StackBlurForCanvas/StackBlurDemo.html
//Yann: added modifications for passing different args to stackBoxBlurImage + wrapped for angulr dependency injection
var mul_table = [ 512,512,456,512,328,456,335,512,405,328,271,456,388,335,292,512,454,405,364,328,298,271,496,456,420,388,360,335,312,292,273,512,482,454,428,405,383,364,345,328,312,298,284,271,259,496,475,456,437,420,404,388,374,360,347,335,323,312,302,292,282,273,265,512,497,482,468,454,441,428,417,405,394,383,373,364,354,345,337,328,320,312,305,298,291,284,278,271,265,259,507,496,485,475,465,456,446,437,428,420,412,404,396,388,381,374,367,360,354,347,341,335,329,323,318,312,307,302,297,292,287,282,278,273,269,265,261,512,505,497,489,482,475,468,461,454,447,441,435,428,422,417,411,405,399,394,389,383,378,373,368,364,359,354,350,345,341,337,332,328,324,320,316,312,309,305,301,298,294,291,287,284,281,278,274,271,268,265,262,259,257,507,501,496,491,485,480,475,470,465,460,456,451,446,442,437,433,428,424,420,416,412,408,404,400,396,392,388,385,381,377,374,370,367,363,360,357,354,350,347,344,341,338,335,332,329,326,323,320,318,315,312,310,307,304,302,299,297,294,292,289,287,285,282,280,278,275,273,271,269,267,265,263,261,259];
var shg_table = [9, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 17,17, 17, 17, 17, 17, 17, 18, 18, 18, 18, 18, 18, 18, 18, 18, 19,19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 20, 20, 20,20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 21,21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21,21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 22, 22, 22, 22, 22, 22,22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22,22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 23,23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23,23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23,23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23,23, 23, 23, 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24 ];

stackBlurImageFn = function stackBlurImage( canvas, radius, blurAlphaChannel)
{
	
	var w = canvas.width;
	var h = canvas.height;
	var context = canvas.getContext('2d');

	if( isNaN(radius) || radius < 1){ return; }

	if( blurAlphaChannel) {
		stackBlurCanvasRGBA( canvas, 0, 0, w, h, radius);
	}else {
		stackBlurCanvasRGB( canvas, 0, 0, w, h, radius);
	}
};


function stackBlurCanvasRGBA( canvas, top_x, top_y, width, height, radius)
{
	if( isNaN(radius) || radius < 1) {
		return;
	}
	radius |= 0;

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height);
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height);
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height);
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e2);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum, a_sum,
		r_out_sum, g_out_sum, b_out_sum, a_out_sum,
		r_in_sum, g_in_sum, b_in_sum, a_in_sum,
		pr, pg, pb, pa, rbs;

	var div = radius + radius + 1;
	var w4 = width << 2;
	var widthMinus1  = width - 1;
	var heightMinus1 = height - 1;
	var radiusPlus1  = radius + 1;
	var sumFactor = radiusPlus1 * ( radiusPlus1 + 1) / 2;

	var stackStart = new BlurStack();
	var stack = stackStart;
	var stackEnd;
	for ( i = 1; i < div; i++)
	{
		stack = stack.next = new BlurStack();
		if( i === radiusPlus1){
			stackEnd = stack;
		}
	}
	stack.next = stackStart;
	var stackIn = null;
	var stackOut = null;

	yw = yi = 0;

	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];

	for ( y = 0; y < height; y++)
	{
		r_in_sum = g_in_sum = b_in_sum = a_in_sum = r_sum = g_sum = b_sum = a_sum = 0;

		r_out_sum = radiusPlus1 * ( pr = pixels[yi]);
		g_out_sum = radiusPlus1 * ( pg = pixels[yi+1]);
		b_out_sum = radiusPlus1 * ( pb = pixels[yi+2]);
		a_out_sum = radiusPlus1 * ( pa = pixels[yi+3]);

		r_sum += sumFactor * pr;
		g_sum += sumFactor * pg;
		b_sum += sumFactor * pb;
		a_sum += sumFactor * pa;

		stack = stackStart;

		for( i = 0; i < radiusPlus1; i++)
		{
			stack.r = pr;
			stack.g = pg;
			stack.b = pb;
			stack.a = pa;
			stack = stack.next;
		}

		for( i = 1; i < radiusPlus1; i++)
		{
			p = yi + (( widthMinus1 < i ? widthMinus1 : i) << 2);
			r_sum += ( stack.r = ( pr = pixels[p])) * ( rbs = radiusPlus1 - i );
			g_sum += ( stack.g = ( pg = pixels[p+1])) * rbs;
			b_sum += ( stack.b = ( pb = pixels[p+2])) * rbs;
			a_sum += ( stack.a = ( pa = pixels[p+3])) * rbs;

			r_in_sum += pr;
			g_in_sum += pg;
			b_in_sum += pb;
			a_in_sum += pa;

			stack = stack.next;
		}


		stackIn = stackStart;
		stackOut = stackEnd;
		for ( x = 0; x < width; x++ )
		{
			pixels[yi+3] = pa = (a_sum * mul_sum) >> shg_sum;
			if( pa !== 0 )
			{
				pa = 255 / pa;
				pixels[yi]   = ((r_sum * mul_sum) >> shg_sum) * pa;
				pixels[yi+1] = ((g_sum * mul_sum) >> shg_sum) * pa;
				pixels[yi+2] = ((b_sum * mul_sum) >> shg_sum) * pa;
			}else {
				pixels[yi] = pixels[yi+1] = pixels[yi+2] = 0;
			}

			r_sum -= r_out_sum;
			g_sum -= g_out_sum;
			b_sum -= b_out_sum;
			a_sum -= a_out_sum;

			r_out_sum -= stackIn.r;
			g_out_sum -= stackIn.g;
			b_out_sum -= stackIn.b;
			a_out_sum -= stackIn.a;

			p =  ( yw + ( ( p = x + radius + 1 ) < widthMinus1 ? p : widthMinus1 ) ) << 2;

			r_in_sum += ( stackIn.r = pixels[p]);
			g_in_sum += ( stackIn.g = pixels[p+1]);
			b_in_sum += ( stackIn.b = pixels[p+2]);
			a_in_sum += ( stackIn.a = pixels[p+3]);

			r_sum += r_in_sum;
			g_sum += g_in_sum;
			b_sum += b_in_sum;
			a_sum += a_in_sum;

			stackIn = stackIn.next;

			r_out_sum += ( pr = stackOut.r );
			g_out_sum += ( pg = stackOut.g );
			b_out_sum += ( pb = stackOut.b );
			a_out_sum += ( pa = stackOut.a );

			r_in_sum -= pr;
			g_in_sum -= pg;
			b_in_sum -= pb;
			a_in_sum -= pa;

			stackOut = stackOut.next;

			yi += 4;
		}
		yw += width;
	}


	for ( x = 0; x < width; x++ )
	{
		g_in_sum = b_in_sum = a_in_sum = r_in_sum = g_sum = b_sum = a_sum = r_sum = 0;

		yi = x << 2;
		r_out_sum = radiusPlus1 * ( pr = pixels[yi]);
		g_out_sum = radiusPlus1 * ( pg = pixels[yi+1]);
		b_out_sum = radiusPlus1 * ( pb = pixels[yi+2]);
		a_out_sum = radiusPlus1 * ( pa = pixels[yi+3]);

		r_sum += sumFactor * pr;
		g_sum += sumFactor * pg;
		b_sum += sumFactor * pb;
		a_sum += sumFactor * pa;

		stack = stackStart;

		for( i = 0; i < radiusPlus1; i++ )
		{
			stack.r = pr;
			stack.g = pg;
			stack.b = pb;
			stack.a = pa;
			stack = stack.next;
		}

		yp = width;

		for( i = 1; i <= radius; i++ )
		{
			yi = ( yp + x ) << 2;

			r_sum += ( stack.r = ( pr = pixels[yi])) * ( rbs = radiusPlus1 - i );
			g_sum += ( stack.g = ( pg = pixels[yi+1])) * rbs;
			b_sum += ( stack.b = ( pb = pixels[yi+2])) * rbs;
			a_sum += ( stack.a = ( pa = pixels[yi+3])) * rbs;

			r_in_sum += pr;
			g_in_sum += pg;
			b_in_sum += pb;
			a_in_sum += pa;

			stack = stack.next;

			if( i < heightMinus1 )
			{
				yp += width;
			}
		}

		yi = x;
		stackIn = stackStart;
		stackOut = stackEnd;
		for ( y = 0; y < height; y++ )
		{
			p = yi << 2;
			pixels[p+3] = pa = (a_sum * mul_sum) >> shg_sum;
			if( pa > 0 )
			{
				pa = 255 / pa;
				pixels[p]   = ((r_sum * mul_sum) >> shg_sum ) * pa;
				pixels[p+1] = ((g_sum * mul_sum) >> shg_sum ) * pa;
				pixels[p+2] = ((b_sum * mul_sum) >> shg_sum ) * pa;
			}else {
				pixels[p] = pixels[p+1] = pixels[p+2] = 0;
			}

			r_sum -= r_out_sum;
			g_sum -= g_out_sum;
			b_sum -= b_out_sum;
			a_sum -= a_out_sum;

			r_out_sum -= stackIn.r;
			g_out_sum -= stackIn.g;
			b_out_sum -= stackIn.b;
			a_out_sum -= stackIn.a;

			p = ( x + (( ( p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1 ) * width )) << 2;

			r_sum += ( r_in_sum += ( stackIn.r = pixels[p]));
			g_sum += ( g_in_sum += ( stackIn.g = pixels[p+1]));
			b_sum += ( b_in_sum += ( stackIn.b = pixels[p+2]));
			a_sum += ( a_in_sum += ( stackIn.a = pixels[p+3]));

			stackIn = stackIn.next;

			r_out_sum += ( pr = stackOut.r );
			g_out_sum += ( pg = stackOut.g );
			b_out_sum += ( pb = stackOut.b );
			a_out_sum += ( pa = stackOut.a );

			r_in_sum -= pr;
			g_in_sum -= pg;
			b_in_sum -= pb;
			a_in_sum -= pa;

			stackOut = stackOut.next;

			yi += width;
		}
	}

	context.putImageData( imageData, top_x, top_y );
}


function stackBlurCanvasRGB( canvas, top_x, top_y, width, height, radius )
{
	if( isNaN(radius) || radius < 1 ) return;
	radius |= 0;

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height );
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height );
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height );
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e2);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum,
		r_out_sum, g_out_sum, b_out_sum,
		r_in_sum, g_in_sum, b_in_sum,
		pr, pg, pb, rbs;

	var div = radius + radius + 1;
	var w4 = width << 2;
	var widthMinus1  = width - 1;
	var heightMinus1 = height - 1;
	var radiusPlus1  = radius + 1;
	var sumFactor = radiusPlus1 * ( radiusPlus1 + 1 ) / 2;

	var stackStart = new BlurStack();
	var stack = stackStart;
	var stackEnd;
	for ( i = 1; i < div; i++ )
	{
		stack = stack.next = new BlurStack();
		if( i === radiusPlus1 ) stackEnd = stack;
	}
	stack.next = stackStart;
	var stackIn = null;
	var stackOut = null;

	yw = yi = 0;

	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];

	for ( y = 0; y < height; y++ )
	{
		r_in_sum = g_in_sum = b_in_sum = r_sum = g_sum = b_sum = 0;

		r_out_sum = radiusPlus1 * ( pr = pixels[yi] );
		g_out_sum = radiusPlus1 * ( pg = pixels[yi+1] );
		b_out_sum = radiusPlus1 * ( pb = pixels[yi+2] );

		r_sum += sumFactor * pr;
		g_sum += sumFactor * pg;
		b_sum += sumFactor * pb;

		stack = stackStart;

		for( i = 0; i < radiusPlus1; i++ )
		{
			stack.r = pr;
			stack.g = pg;
			stack.b = pb;
			stack = stack.next;
		}

		for( i = 1; i < radiusPlus1; i++ )
		{
			p = yi + (( widthMinus1 < i ? widthMinus1 : i ) << 2 );
			r_sum += ( stack.r = ( pr = pixels[p])) * ( rbs = radiusPlus1 - i );
			g_sum += ( stack.g = ( pg = pixels[p+1])) * rbs;
			b_sum += ( stack.b = ( pb = pixels[p+2])) * rbs;

			r_in_sum += pr;
			g_in_sum += pg;
			b_in_sum += pb;

			stack = stack.next;
		}


		stackIn = stackStart;
		stackOut = stackEnd;
		for ( x = 0; x < width; x++ )
		{
			pixels[yi]   = (r_sum * mul_sum) >> shg_sum;
			pixels[yi+1] = (g_sum * mul_sum) >> shg_sum;
			pixels[yi+2] = (b_sum * mul_sum) >> shg_sum;

			r_sum -= r_out_sum;
			g_sum -= g_out_sum;
			b_sum -= b_out_sum;

			r_out_sum -= stackIn.r;
			g_out_sum -= stackIn.g;
			b_out_sum -= stackIn.b;

			p =  ( yw + ( ( p = x + radius + 1 ) < widthMinus1 ? p : widthMinus1 ) ) << 2;

			r_in_sum += ( stackIn.r = pixels[p]);
			g_in_sum += ( stackIn.g = pixels[p+1]);
			b_in_sum += ( stackIn.b = pixels[p+2]);

			r_sum += r_in_sum;
			g_sum += g_in_sum;
			b_sum += b_in_sum;

			stackIn = stackIn.next;

			r_out_sum += ( pr = stackOut.r );
			g_out_sum += ( pg = stackOut.g );
			b_out_sum += ( pb = stackOut.b );

			r_in_sum -= pr;
			g_in_sum -= pg;
			b_in_sum -= pb;

			stackOut = stackOut.next;

			yi += 4;
		}
		yw += width;
	}


	for ( x = 0; x < width; x++ )
	{
		g_in_sum = b_in_sum = r_in_sum = g_sum = b_sum = r_sum = 0;

		yi = x << 2;
		r_out_sum = radiusPlus1 * ( pr = pixels[yi]);
		g_out_sum = radiusPlus1 * ( pg = pixels[yi+1]);
		b_out_sum = radiusPlus1 * ( pb = pixels[yi+2]);

		r_sum += sumFactor * pr;
		g_sum += sumFactor * pg;
		b_sum += sumFactor * pb;

		stack = stackStart;

		for( i = 0; i < radiusPlus1; i++ )
		{
			stack.r = pr;
			stack.g = pg;
			stack.b = pb;
			stack = stack.next;
		}

		yp = width;

		for( i = 1; i <= radius; i++ )
		{
			yi = ( yp + x ) << 2;

			r_sum += ( stack.r = ( pr = pixels[yi])) * ( rbs = radiusPlus1 - i );
			g_sum += ( stack.g = ( pg = pixels[yi+1])) * rbs;
			b_sum += ( stack.b = ( pb = pixels[yi+2])) * rbs;

			r_in_sum += pr;
			g_in_sum += pg;
			b_in_sum += pb;

			stack = stack.next;

			if( i < heightMinus1 )
			{
				yp += width;
			}
		}

		yi = x;
		stackIn = stackStart;
		stackOut = stackEnd;
		for ( y = 0; y < height; y++ )
		{
			p = yi << 2;
			pixels[p]   = (r_sum * mul_sum) >> shg_sum;
			pixels[p+1] = (g_sum * mul_sum) >> shg_sum;
			pixels[p+2] = (b_sum * mul_sum) >> shg_sum;

			r_sum -= r_out_sum;
			g_sum -= g_out_sum;
			b_sum -= b_out_sum;

			r_out_sum -= stackIn.r;
			g_out_sum -= stackIn.g;
			b_out_sum -= stackIn.b;

			p = ( x + (( ( p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1 ) * width )) << 2;

			r_sum += ( r_in_sum += ( stackIn.r = pixels[p]));
			g_sum += ( g_in_sum += ( stackIn.g = pixels[p+1]));
			b_sum += ( b_in_sum += ( stackIn.b = pixels[p+2]));

			stackIn = stackIn.next;

			r_out_sum += ( pr = stackOut.r );
			g_out_sum += ( pg = stackOut.g );
			b_out_sum += ( pb = stackOut.b );

			r_in_sum -= pr;
			g_in_sum -= pg;
			b_in_sum -= pb;

			stackOut = stackOut.next;

			yi += width;
		}
	}

	context.putImageData( imageData, top_x, top_y );

}

function BlurStack()
{
	this.r = 0;
	this.g = 0;
	this.b = 0;
	this.a = 0;
	this.next = null;
}
})();
	

(function(){
//http://www.quasimondo.com/BoxBlurForCanvas/StackBoxBlur.js
//http://www.quasimondo.com/BoxBlurForCanvas/FastBlur2Demo.html
//Yann: added modifications for passing different args to stackBoxBlurImage + wrapped for angulr dependency injection
var mul_table = [ 1,171,205,293,57,373,79,137,241,27,391,357,41,19,283,265,497,469,443,421,25,191,365,349,335,161,155,149,9,278,269,261,505,245,475,231,449,437,213,415,405,395,193,377,369,361,353,345,169,331,325,319,313,307,301,37,145,285,281,69,271,267,263,259,509,501,493,243,479,118,465,459,113,446,55,435,429,423,209,413,51,403,199,393,97,3,379,375,371,367,363,359,355,351,347,43,85,337,333,165,327,323,5,317,157,311,77,305,303,75,297,294,73,289,287,71,141,279,277,275,68,135,67,133,33,262,260,129,511,507,503,499,495,491,61,121,481,477,237,235,467,232,115,457,227,451,7,445,221,439,218,433,215,427,425,211,419,417,207,411,409,203,202,401,399,396,197,49,389,387,385,383,95,189,47,187,93,185,23,183,91,181,45,179,89,177,11,175,87,173,345,343,341,339,337,21,167,83,331,329,327,163,81,323,321,319,159,79,315,313,39,155,309,307,153,305,303,151,75,299,149,37,295,147,73,291,145,289,287,143,285,71,141,281,35,279,139,69,275,137,273,17,271,135,269,267,133,265,33,263,131,261,130,259,129,257,1];
var shg_table = [0,9,10,11,9,12,10,11,12,9,13,13,10,9,13,13,14,14,14,14,10,13,14,14,14,13,13,13,9,14,14,14,15,14,15,14,15,15,14,15,15,15,14,15,15,15,15,15,14,15,15,15,15,15,15,12,14,15,15,13,15,15,15,15,16,16,16,15,16,14,16,16,14,16,13,16,16,16,15,16,13,16,15,16,14,9,16,16,16,16,16,16,16,16,16,13,14,16,16,15,16,16,10,16,15,16,14,16,16,14,16,16,14,16,16,14,15,16,16,16,14,15,14,15,13,16,16,15,17,17,17,17,17,17,14,15,17,17,16,16,17,16,15,17,16,17,11,17,16,17,16,17,16,17,17,16,17,17,16,17,17,16,16,17,17,17,16,14,17,17,17,17,15,16,14,16,15,16,13,16,15,16,14,16,15,16,12,16,15,16,17,17,17,17,17,13,16,15,17,17,17,16,15,17,17,17,16,15,17,17,14,16,17,17,16,17,17,16,15,17,16,14,17,16,15,17,16,17,17,16,17,15,16,17,14,17,16,15,17,16,17,13,17,16,17,17,16,17,14,17,16,17,16,17,16,17,9];

stackBoxBlurImageFn = function stackBoxBlurImage( canvas, radius, blurAlphaChannel, iterations )
{
	iterations = iterations || 1;
	var w = canvas.width;
	var h = canvas.height;
	var context = canvas.getContext('2d');
	
	if( isNaN(radius) || radius < 1 ){ return; }
	
	if( blurAlphaChannel ){
		stackBoxBlurCanvasRGBA( canvas, 0, 0, w, h, radius, iterations );
	}else{
		stackBoxBlurCanvasRGB( canvas, 0, 0, w, h, radius, iterations );
	}
	
	return canvas;
};


function stackBoxBlurCanvasRGBA( canvas, top_x, top_y, width, height, radius, iterations )
{
	if( isNaN(radius) || radius < 1 ){ return; }
	radius |= 0;

	if( isNaN(iterations) ){ iterations = 1; }
	iterations |= 0;
	if( iterations > 3 ){ iterations = 3; }
	if( iterations < 1 ){ iterations = 1; }

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height );
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height );
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height );
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum, a_sum,
		r_out_sum, g_out_sum, b_out_sum, a_out_sum,
		r_in_sum, g_in_sum, b_in_sum, a_in_sum,
		pr, pg, pb, pa, rbs;

	var div = radius + radius + 1;
	var w4 = width << 2;
	var widthMinus1  = width - 1;
	var heightMinus1 = height - 1;
	var radiusPlus1  = radius + 1;

	var stackStart = new BlurStack();

	var stack = stackStart;
	for ( i = 1; i < div; i++ )
	{
		stack = stack.next = new BlurStack();
		if( i === radiusPlus1 ){ var stackEnd = stack; }
	}
	stack.next = stackStart;
	var stackIn = null;



	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];
	while ( iterations-- > 0 ) {
		yw = yi = 0;
		for ( y = height; --y > -1; )
		{
			r_sum = radiusPlus1 * ( pr = pixels[yi] );
			g_sum = radiusPlus1 * ( pg = pixels[yi+1] );
			b_sum = radiusPlus1 * ( pb = pixels[yi+2] );
			a_sum = radiusPlus1 * ( pa = pixels[yi+3] );

			stack = stackStart;

			for( i = radiusPlus1; --i > -1; )
			{
				stack.r = pr;
				stack.g = pg;
				stack.b = pb;
				stack.a = pa;
				stack = stack.next;
			}

			for( i = 1; i < radiusPlus1; i++ )
			{
				p = yi + (( widthMinus1 < i ? widthMinus1 : i ) << 2 );
				r_sum += ( stack.r = pixels[p]);
				g_sum += ( stack.g = pixels[p+1]);
				b_sum += ( stack.b = pixels[p+2]);
				a_sum += ( stack.a = pixels[p+3]);

				stack = stack.next;
			}

			stackIn = stackStart;
			for ( x = 0; x < width; x++ )
			{
				pixels[yi++] = (r_sum * mul_sum) >>> shg_sum;
				pixels[yi++] = (g_sum * mul_sum) >>> shg_sum;
				pixels[yi++] = (b_sum * mul_sum) >>> shg_sum;
				pixels[yi++] = (a_sum * mul_sum) >>> shg_sum;

				p =  ( yw + ( ( p = x + radius + 1 ) < widthMinus1 ? p : widthMinus1 ) ) << 2;

				r_sum -= stackIn.r - ( stackIn.r = pixels[p]);
				g_sum -= stackIn.g - ( stackIn.g = pixels[p+1]);
				b_sum -= stackIn.b - ( stackIn.b = pixels[p+2]);
				a_sum -= stackIn.a - ( stackIn.a = pixels[p+3]);

				stackIn = stackIn.next;

			}
			yw += width;
		}


		for ( x = 0; x < width; x++ )
		{
			yi = x << 2;

			r_sum = radiusPlus1 * ( pr = pixels[yi]);
			g_sum = radiusPlus1 * ( pg = pixels[yi+1]);
			b_sum = radiusPlus1 * ( pb = pixels[yi+2]);
			a_sum = radiusPlus1 * ( pa = pixels[yi+3]);

			stack = stackStart;

			for( i = 0; i < radiusPlus1; i++ )
			{
				stack.r = pr;
				stack.g = pg;
				stack.b = pb;
				stack.a = pa;
				stack = stack.next;
			}

			yp = width;

			for( i = 1; i <= radius; i++ )
			{
				yi = ( yp + x ) << 2;

				r_sum += ( stack.r = pixels[yi]);
				g_sum += ( stack.g = pixels[yi+1]);
				b_sum += ( stack.b = pixels[yi+2]);
				a_sum += ( stack.a = pixels[yi+3]);

				stack = stack.next;

				if( i < heightMinus1 )
				{
					yp += width;
				}
			}

			yi = x;
			stackIn = stackStart;
			for ( y = 0; y < height; y++ )
			{
				p = yi << 2;
				pixels[p+3] = pa =(a_sum * mul_sum) >>> shg_sum;
				if( pa > 0 )
				{
					pa = 255 / pa;
					pixels[p]   = ((r_sum * mul_sum) >>> shg_sum ) * pa;
					pixels[p+1] = ((g_sum * mul_sum) >>> shg_sum ) * pa;
					pixels[p+2] = ((b_sum * mul_sum) >>> shg_sum ) * pa;
				}else {
					pixels[p] = pixels[p+1] = pixels[p+2] = 0;
				}

				p = ( x + (( ( p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1 ) * width )) << 2;

				r_sum -= stackIn.r - ( stackIn.r = pixels[p]);
				g_sum -= stackIn.g - ( stackIn.g = pixels[p+1]);
				b_sum -= stackIn.b - ( stackIn.b = pixels[p+2]);
				a_sum -= stackIn.a - ( stackIn.a = pixels[p+3]);

				stackIn = stackIn.next;

				yi += width;
			}
		}
	}
	context.putImageData( imageData, top_x, top_y );

}


function stackBoxBlurCanvasRGB( canvas, top_x, top_y, width, height, radius, iterations )
{
	if( isNaN(radius) || radius < 1 ){ return; }
	radius |= 0;

	if( isNaN(iterations) ){ iterations = 1; }
	iterations |= 0;
	if( iterations > 3 ){ iterations = 3; }
	if( iterations < 1 ){ iterations = 1; }

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height );
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height );
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height );
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e2);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum,
		r_out_sum, g_out_sum, b_out_sum,
		r_in_sum, g_in_sum, b_in_sum,
		pr, pg, pb, rbs;

	var div = radius + radius + 1;
	var w4 = width << 2;
	var widthMinus1  = width - 1;
	var heightMinus1 = height - 1;
	var radiusPlus1  = radius + 1;

	var stackStart = new BlurStack();
	var stack = stackStart;
	for ( i = 1; i < div; i++ )
	{
		stack = stack.next = new BlurStack();
		if( i === radiusPlus1 ){ var stackEnd = stack; }
	}
	stack.next = stackStart;
	var stackIn = null;



	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];

	while ( iterations-- > 0 ) {
		yw = yi = 0;

		for ( y = height; --y >-1; )
		{
			r_sum = radiusPlus1 * ( pr = pixels[yi] );
			g_sum = radiusPlus1 * ( pg = pixels[yi+1] );
			b_sum = radiusPlus1 * ( pb = pixels[yi+2] );

			stack = stackStart;

			for( i = radiusPlus1; --i > -1; )
			{
				stack.r = pr;
				stack.g = pg;
				stack.b = pb;
				stack = stack.next;
			}

			for( i = 1; i < radiusPlus1; i++ )
			{
				p = yi + (( widthMinus1 < i ? widthMinus1 : i ) << 2 );
				r_sum += ( stack.r = pixels[p++]);
				g_sum += ( stack.g = pixels[p++]);
				b_sum += ( stack.b = pixels[p]);

				stack = stack.next;
			}

			stackIn = stackStart;
			for ( x = 0; x < width; x++ )
			{
				pixels[yi++] = (r_sum * mul_sum) >>> shg_sum;
				pixels[yi++] = (g_sum * mul_sum) >>> shg_sum;
				pixels[yi++] = (b_sum * mul_sum) >>> shg_sum;
				yi++;

				p =  ( yw + ( ( p = x + radius + 1 ) < widthMinus1 ? p : widthMinus1 ) ) << 2;

				r_sum -= stackIn.r - ( stackIn.r = pixels[p++]);
				g_sum -= stackIn.g - ( stackIn.g = pixels[p++]);
				b_sum -= stackIn.b - ( stackIn.b = pixels[p]);

				stackIn = stackIn.next;
			}
			yw += width;
		}


		for ( x = 0; x < width; x++ )
		{
			yi = x << 2;

			r_sum = radiusPlus1 * ( pr = pixels[yi++]);
			g_sum = radiusPlus1 * ( pg = pixels[yi++]);
			b_sum = radiusPlus1 * ( pb = pixels[yi]);

			stack = stackStart;

			for( i = 0; i < radiusPlus1; i++ )
			{
				stack.r = pr;
				stack.g = pg;
				stack.b = pb;
				stack = stack.next;
			}

			yp = width;

			for( i = 1; i <= radius; i++ )
			{
				yi = ( yp + x ) << 2;

				r_sum += ( stack.r = pixels[yi++]);
				g_sum += ( stack.g = pixels[yi++]);
				b_sum += ( stack.b = pixels[yi]);

				stack = stack.next;

				if( i < heightMinus1 ){ yp += width; }
			}

			yi = x;
			stackIn = stackStart;
			for ( y = 0; y < height; y++ )
			{
				p = yi << 2;
				pixels[p]   = (r_sum * mul_sum) >>> shg_sum;
				pixels[p+1] = (g_sum * mul_sum) >>> shg_sum;
				pixels[p+2] = (b_sum * mul_sum) >>> shg_sum;

				p = ( x + (( ( p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1 ) * width )) << 2;

				r_sum -= stackIn.r - ( stackIn.r = pixels[p]);
				g_sum -= stackIn.g - ( stackIn.g = pixels[p+1]);
				b_sum -= stackIn.b - ( stackIn.b = pixels[p+2]);

				stackIn = stackIn.next;

				yi += width;
			}
		}
	}
	context.putImageData( imageData, top_x, top_y );

}

function BlurStack()
{
	this.r = 0;
	this.g = 0;
	this.b = 0;
	this.a = 0;
	this.next = null;
}
})();

	
(function(){

//http://www.quasimondo.com/BoxBlurForCanvas/FastBlur.js
//http://www.quasimondo.com/BoxBlurForCanvas/FastBlurDemo.html
var mul_table = [ 1,57,41,21,203,34,97,73,227,91,149,62,105,45,39,137,241,107,3,173,39,71,65,238,219,101,187,87,81,151,141,133,249,117,221,209,197,187,177,169,5,153,73,139,133,127,243,233,223,107,103,99,191,23,177,171,165,159,77,149,9,139,135,131,253,245,119,231,224,109,211,103,25,195,189,23,45,175,171,83,81,79,155,151,147,9,141,137,67,131,129,251,123,30,235,115,113,221,217,53,13,51,50,49,193,189,185,91,179,175,43,169,83,163,5,79,155,19,75,147,145,143,35,69,17,67,33,65,255,251,247,243,239,59,29,229,113,111,219,27,213,105,207,51,201,199,49,193,191,47,93,183,181,179,11,87,43,85,167,165,163,161,159,157,155,77,19,75,37,73,145,143,141,35,138,137,135,67,33,131,129,255,63,250,247,61,121,239,237,117,29,229,227,225,111,55,109,216,213,211,209,207,205,203,201,199,197,195,193,48,190,47,93,185,183,181,179,178,176,175,173,171,85,21,167,165,41,163,161,5,79,157,78,154,153,19,75,149,74,147,73,144,143,71,141,140,139,137,17,135,134,133,66,131,65,129,1];
var shg_table = [0,9,10,10,14,12,14,14,16,15,16,15,16,15,15,17,18,17,12,18,16,17,17,19,19,18,19,18,18,19,19,19,20,19,20,20,20,20,20,20,15,20,19,20,20,20,21,21,21,20,20,20,21,18,21,21,21,21,20,21,17,21,21,21,22,22,21,22,22,21,22,21,19,22,22,19,20,22,22,21,21,21,22,22,22,18,22,22,21,22,22,23,22,20,23,22,22,23,23,21,19,21,21,21,23,23,23,22,23,23,21,23,22,23,18,22,23,20,22,23,23,23,21,22,20,22,21,22,24,24,24,24,24,22,21,24,23,23,24,21,24,23,24,22,24,24,22,24,24,22,23,24,24,24,20,23,22,23,24,24,24,24,24,24,24,23,21,23,22,23,24,24,24,22,24,24,24,23,22,24,24,25,23,25,25,23,24,25,25,24,22,25,25,25,24,23,24,25,25,25,25,25,25,25,25,25,25,25,25,23,25,23,24,25,25,25,25,25,25,25,25,25,24,22,25,25,23,25,25,20,24,25,24,25,25,22,24,25,24,25,24,25,25,24,25,25,25,25,22,25,25,25,24,25,24,25,18];


boxBlurImageFn = function boxBlurImage( canvas, radius, blurAlphaChannel, iterations ){
	var w = canvas.width;
	var h = canvas.height;
	var context = canvas.getContext('2d');
	
	if( isNaN(radius) || radius < 1 ) return;

	if( blurAlphaChannel )
	{
		boxBlurCanvasRGBA( canvas, 0, 0, w, h, radius, iterations );
	}else {
		boxBlurCanvasRGB( canvas, 0, 0, w, h, radius, iterations );
	}
};


function boxBlurCanvasRGBA( canvas, top_x, top_y, width, height, radius, iterations ){
	if( isNaN(radius) || radius < 1 ) return;

	radius |= 0;

	if( isNaN(iterations) ) iterations = 1;
	iterations |= 0;
	if( iterations > 3 ) iterations = 3;
	if( iterations < 1 ) iterations = 1;

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height );
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height );
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height );
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e2);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var rsum,gsum,bsum,asum,x,y,i,p,p1,p2,yp,yi,yw,idx,pa;
	var wm = width - 1;
	var hm = height - 1;
	var wh = width * height;
	var rad1 = radius + 1;

	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];

	var r = [];
	var g = [];
	var b = [];
	var a = [];

	var vmin = [];
	var vmax = [];

	while ( iterations-- > 0 ){
		yw = yi = 0;

		for ( y=0; y < height; y++ ){
			rsum = pixels[yw]   * rad1;
			gsum = pixels[yw+1] * rad1;
			bsum = pixels[yw+2] * rad1;
			asum = pixels[yw+3] * rad1;


			for( i = 1; i <= radius; i++ ){
				p = yw + (((i > wm ? wm : i )) << 2 );
				rsum += pixels[p++];
				gsum += pixels[p++];
				bsum += pixels[p++];
				asum += pixels[p];
			}

			for ( x = 0; x < width; x++ ) {
				r[yi] = rsum;
				g[yi] = gsum;
				b[yi] = bsum;
				a[yi] = asum;

				if( y===0) {
					vmin[x] = ( ( p = x + rad1) < wm ? p : wm ) << 2;
					vmax[x] = ( ( p = x - radius) > 0 ? p << 2 : 0 );
				}

				p1 = yw + vmin[x];
				p2 = yw + vmax[x];

				rsum += pixels[p1++] - pixels[p2++];
				gsum += pixels[p1++] - pixels[p2++];
				bsum += pixels[p1++] - pixels[p2++];
				asum += pixels[p1]   - pixels[p2];

				yi++;
			}
			yw += ( width << 2 );
		}

		for ( x = 0; x < width; x++ ) {
			yp = x;
			rsum = r[yp] * rad1;
			gsum = g[yp] * rad1;
			bsum = b[yp] * rad1;
			asum = a[yp] * rad1;

			for( i = 1; i <= radius; i++ ) {
				yp += ( i > hm ? 0 : width );
				rsum += r[yp];
				gsum += g[yp];
				bsum += b[yp];
				asum += a[yp];
			}

			yi = x << 2;
			for ( y = 0; y < height; y++) {

				pixels[yi+3] = pa = (asum * mul_sum) >>> shg_sum;
				if( pa > 0 )
				{
					pa = 255 / pa;
					pixels[yi]   = ((rsum * mul_sum) >>> shg_sum) * pa;
					pixels[yi+1] = ((gsum * mul_sum) >>> shg_sum) * pa;
					pixels[yi+2] = ((bsum * mul_sum) >>> shg_sum) * pa;
				}else {
					pixels[yi] = pixels[yi+1] = pixels[yi+2] = 0;
				}
				if( x === 0 ) {
					vmin[y] = ( ( p = y + rad1) < hm ? p : hm ) * width;
					vmax[y] = ( ( p = y - radius) > 0 ? p * width : 0 );
				}

				p1 = x + vmin[y];
				p2 = x + vmax[y];

				rsum += r[p1] - r[p2];
				gsum += g[p1] - g[p2];
				bsum += b[p1] - b[p2];
				asum += a[p1] - a[p2];

				yi += width << 2;
			}
		}
	}

	context.putImageData( imageData, top_x, top_y );

}

function boxBlurCanvasRGB( canvas, top_x, top_y, width, height, radius, iterations ){
	if( isNaN(radius) || radius < 1 ) return;

	radius |= 0;

	if( isNaN(iterations) ) iterations = 1;
	iterations |= 0;
	if( iterations > 3 ) iterations = 3;
	if( iterations < 1 ) iterations = 1;

	var context = canvas.getContext('2d');
	var imageData;

	try {
		try {
			imageData = context.getImageData( top_x, top_y, width, height );
		} catch(e) {

			// NOTE: this part is supposedly only needed if you want to work with local files
			// so it might be okay to remove the whole try/catch block and just use
			// imageData = context.getImageData( top_x, top_y, width, height );
			try {
				netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
				imageData = context.getImageData( top_x, top_y, width, height );
			} catch(e2) {
				alert('Cannot access local image');
				throw new Error('unable to access local image data: ' + e2);
			}
		}
	} catch(e) {
		alert('Cannot access image');
		throw new Error('unable to access image data: ' + e);
	}

	var pixels = imageData.data;

	var rsum,gsum,bsum,asum,x,y,i,p,p1,p2,yp,yi,yw,idx;
	var wm = width - 1;
	var hm = height - 1;
	var wh = width * height;
	var rad1 = radius + 1;

	var r = [];
	var g = [];
	var b = [];

	var mul_sum = mul_table[radius];
	var shg_sum = shg_table[radius];

	var vmin = [];
	var vmax = [];

	while ( iterations-- > 0 ){
		yw = yi = 0;

		for ( y=0; y < height; y++ ){
			rsum = pixels[yw]   * rad1;
			gsum = pixels[yw+1] * rad1;
			bsum = pixels[yw+2] * rad1;

			for( i = 1; i <= radius; i++ ){
				p = yw + (((i > wm ? wm : i )) << 2 );
				rsum += pixels[p++];
				gsum += pixels[p++];
				bsum += pixels[p++];
			}

			for ( x = 0; x < width; x++ ){
				r[yi] = rsum;
				g[yi] = gsum;
				b[yi] = bsum;

				if( y===0) {
					vmin[x] = ( ( p = x + rad1) < wm ? p : wm ) << 2;
					vmax[x] = ( ( p = x - radius) > 0 ? p << 2 : 0 );
				}

				p1 = yw + vmin[x];
				p2 = yw + vmax[x];

				rsum += pixels[p1++] - pixels[p2++];
				gsum += pixels[p1++] - pixels[p2++];
				bsum += pixels[p1++] - pixels[p2++];

				yi++;
			}
			yw += ( width << 2 );
		}

		for ( x = 0; x < width; x++ ){
			yp = x;
			rsum = r[yp] * rad1;
			gsum = g[yp] * rad1;
			bsum = b[yp] * rad1;

			for( i = 1; i <= radius; i++ ){
				yp += ( i > hm ? 0 : width );
				rsum += r[yp];
				gsum += g[yp];
				bsum += b[yp];
			}

			yi = x << 2;
			for ( y = 0; y < height; y++){
				pixels[yi]   = (rsum * mul_sum) >>> shg_sum;
				pixels[yi+1] = (gsum * mul_sum) >>> shg_sum;
				pixels[yi+2] = (bsum * mul_sum) >>> shg_sum;

				if( x === 0 ) {
					vmin[y] = ( ( p = y + rad1) < hm ? p : hm ) * width;
					vmax[y] = ( ( p = y - radius) > 0 ? p * width : 0 );
				}

				p1 = x + vmin[y];
				p2 = x + vmax[y];

				rsum += r[p1] - r[p2];
				gsum += g[p1] - g[p2];
				bsum += b[p1] - b[p2];

				yi += width << 2;
			}
		}
	}
	context.putImageData( imageData, top_x, top_y );

}
})();
})();
angular.module('fangaApp.misc').service('drawUtils',
["canvasBlurUtils", function(canvasBlurUtils){
	//duplicate of Bubble.kinds, avoids the model dependency for external use
	var BUBBLE_KINDS = {
		RECTANGLE: 'rectangle',
		POLYGON: 'polygon',
		POINT: 'point'
	};
	//duplicate of BubbleStyle.BACKGROUND_TYPES
	var BACKGROUND_TYPES = {
		COLOR_FILL : 'colorFill',
		BLUR : 'blur'
	};
	
	var blurCanvas = document.createElement('canvas');
	var blurCtx = blurCanvas.getContext('2d');
	var blurCanvas2 = document.createElement('canvas');
	var blurCtx2 = blurCanvas2.getContext('2d');
	
	/*
	//Debug Blur - uncomment to display blurCanvas
	$('body').append(blurCanvas);
	$(blurCanvas).css({
		bottom: '0',
		left: '0',
		background: 'white',
		position: 'absolute',
		display: 'block'
	});
	*/
	
	function clearFrame(frame, ctx){
		var kind = frame.kind || BUBBLE_KINDS.RECTANGLE;
		
		if(kind == BUBBLE_KINDS.RECTANGLE){
			ctx.clearRect(
				frame.rect.x,
				frame.rect.y,
				frame.rect.width,
				frame.rect.height
			);
		}else{
			ctx.save();
			ctx.globalCompositeOperation = 'destination-out';
			ctx.fillStyle = 'black';
			if(kind == BUBBLE_KINDS.POLYGON){
				drawPolygon(frame, frame.rect, ctx);
			}else if(kind == BUBBLE_KINDS.POINT){
				ctx.beginPath();
				ctx.arc(frame.rect.x, frame.rect.y, frame.radius, 0, 2 * Math.PI, false);
			}else{
				throw new Error();
			}
			ctx.fill();
			ctx.restore();
		}
	}

	function drawBubble(bubble, rect, strokeOffset, ctx){
		strokeOffset = strokeOffset || 0;

		var kind = bubble.kind;
		if(typeof bubble.text == 'string' && bubble.kind === BUBBLE_KINDS.POLYGON){
			//bubble is actually a bubble translation
			kind = BUBBLE_KINDS.RECTANGLE;
		}

		if(kind == BUBBLE_KINDS.RECTANGLE){
			ctx.beginPath();
			ctx.rect(
				rect.x + strokeOffset,
				rect.y + strokeOffset,
				rect.width - 2 * strokeOffset,
				rect.height - 2 * strokeOffset
			);
		}else if(kind == BUBBLE_KINDS.POLYGON){
			drawPolygon(bubble, rect, ctx);
		}else if(kind == BUBBLE_KINDS.POINT){
			ctx.beginPath();
			ctx.arc(rect.x, rect.y, bubble.radius, 0, 2 * Math.PI, false);
		}
	}

	function drawBubbleBackground(bubble, panel, spread, rect, spreadCanvas, ctx){
		//NOTE: no bubbleStyle for annotations
		if(!bubble.bubbleStyle){
			return;
		}

		if(bubble.bubbleStyle.backgroundType == BACKGROUND_TYPES.COLOR_FILL){
			//Draw the bubble shape over the page
			ctx.fillStyle = bubble.bubbleStyle.backgroundColor;
			drawBubble(bubble, rect, 0, ctx);
			ctx.fill();
		}else if(bubble.bubbleStyle.backgroundType == BACKGROUND_TYPES.BLUR){
			//NOTE: the blur has to be done at 1:1 scale even though it is slower, otherwise blur intensity varies with dispaly zoom
			//NOTE2: if we need to improve performance, we could look into caching the blur result for each bubble (in mainCanvas.directive)

			var blurRadius = bubble.bubbleStyle.blurRadius;
			var blurBleedDistance = bubble.bubbleStyle.blurBleedDistance || 0;
			var blurColorAlpha = bubble.bubbleStyle.blurColorAlpha || 0;

			var blurMargin =  blurBleedDistance + blurRadius;
			var bubbleRect = {
				x: (panel.x + bubble.x * panel.width) * spread.width,
				y: (panel.y + bubble.y * panel.height) * spread.height,
				width: bubble.width * panel.width * spread.width,
				height: bubble.height * panel.height * spread.height
			};
			var drawRect = {};
			drawRect.x = bubbleRect.x - blurMargin;
			drawRect.y = bubbleRect.y - blurMargin;
			drawRect.width = bubbleRect.width + 2 * blurMargin;
			drawRect.height = bubbleRect.height + 2 * blurMargin;

			blurCanvas.width = drawRect.width;
			blurCanvas.height = drawRect.height;

			//Draw the part of the image that is behind the bubble frame in the blur canvas
			blurCtx.drawImage(spreadCanvas,
				drawRect.x, drawRect.y, drawRect.width, drawRect.height,
				0, 0, drawRect.width, drawRect.height
			);
			if(blurColorAlpha > 0){
				//apply bubble background color
				blurCtx.save();
				blurCtx.fillStyle = bubble.bubbleStyle.backgroundColor;
				blurCtx.globalAlpha = blurColorAlpha;
				blurCtx.fillRect(0, 0, blurCanvas.width, blurCanvas.height);
				blurCtx.restore();
			}
			//Apply blur on the image
			canvasBlurUtils.blur(blurCanvas, blurRadius, false, 2);

			//Draw the shape that will be used to cut out the bubble shape in the blur canvas
			var numPixelsBleed = blurBleedDistance;
			if(numPixelsBleed <= 0){
				//Fast path: draw the shape in the blurCanvas directly
				blurCtx.save();
				blurCtx.globalCompositeOperation = 'destination-in';
				blurCtx.fillStyle = 'red';
				blurCtx.strokeStyle = 'red';
				blurCtx.translate(-drawRect.x, -drawRect.y);
				drawBubble(bubble, bubbleRect, 0, blurCtx);
				blurCtx.fill();
				blurCtx.restore();
			}else{
				//Slower path, we need an other off screen canvas to draw the shape + a glow
				blurCanvas2.width = blurCanvas.width;
				blurCanvas2.height = blurCanvas.height;

				//We make a glow-like on the shape to soften edges with a gradient like opacity
				//We achieve this by stacking draw calls with different lineWidths
				blurCtx2.save();
				blurCtx2.fillStyle = 'red';
				blurCtx2.strokeStyle = 'red';
				blurCtx2.lineJoin = 'round';
				blurCtx2.translate(-drawRect.x, -drawRect.y);
				drawBubble(bubble, bubbleRect, 0, blurCtx2);

				for(var i = 1 ; i <= numPixelsBleed; i++){
					//We want the draw alpha to be i/n starting from the edge
					//However we stack draws going from wide to narrow, so we have to compute alpha blending
					// http://www.w3.org/TR/2014/CR-compositing-1-20140220/#generalformula (5.1.1)
					// alphaResult = alphaSource + alphaBackdrop * (1 - alphaSource) = alphaSource * (1-alphaBackdrop) + alphaBackdrop
					// => alphaSource = (alphaResult - alphaBackdrop) / (1-alphaBackdrop)
					var prevAlpha = (i - 1) / (numPixelsBleed + 1);
					var wantedAlpha = i / (numPixelsBleed + 1); //from 1/(N+1) to n/(n+1) not interested in drawing opaque or transparent.
					var drawAlpha = (wantedAlpha - prevAlpha) / (1 - prevAlpha);
					blurCtx2.globalAlpha = drawAlpha;
					blurCtx2.lineWidth = (numPixelsBleed - i + 1) * 2;		// from numPixelsBleed to 1
					blurCtx2.stroke();

					//console.log(drawAlpha, (numPixelsBleed-i+1));
				}
				blurCtx2.globalAlpha = 1;
				blurCtx2.fill();
				blurCtx2.restore();

				//draw this shape in the blurCanvas, acting as its new alpha layer
				blurCtx.save();
				blurCtx.globalCompositeOperation = 'destination-in';
				blurCtx.drawImage(blurCanvas2, 0, 0);
				blurCtx.restore();
			}

			//Draw blurred image in the page
			var scale = rect.width / bubbleRect.width;
			ctx.drawImage(blurCanvas,
				0, 0, blurCanvas.width, blurCanvas.height,
				rect.x - blurMargin * scale,
				rect.y - blurMargin * scale,
				drawRect.width * scale,
				drawRect.height * scale);

			/*
			var blurRadius = bubble.bubbleStyle.blurRadius;
			var blurMargin =  100;
			blurCanvas.width = rect.width + 2*blurMargin;
			blurCanvas.height = rect.height + 2*blurMargin;
			//Draw the part of the image that is behind the bubble shape in the blur canvas
			blurCtx.drawImage(spreadCanvas,
				(panel.x + bubble.x*panel.width)*spread.width,
				(panel.y + bubble.y*panel.height)*spread.height,
				bubble.width*panel.width*spread.width,
				bubble.height*panel.height*spread.height,
				blurMargin, blurMargin, rect.width, rect.height);

			if(bubble.kind === BUBBLE_KINDS.POLYGON){
				//For polygon, we only draw the part of the image under the bubble
				blurCtx.save();
				blurCtx.globalCompositeOperation = 'destination-in';
				blurCtx.fillStyle = 'red';
				blurCtx.translate(-rect.x+blurMargin, -rect.y+blurMargin);
				drawBubble(bubble, rect, 0, blurCtx);
				blurCtx.fill();
				blurCtx.restore();
			}
			//Apply blur
			canvasBlurUtils.blur(blurCanvas, blurRadius, true, 2);

			//Draw blurred image in the page
			ctx.save();
			ctx.drawImage(blurCanvas,rect.x-blurMargin, rect.y-blurMargin);
			ctx.restore();
			*/
		}
	}

	function drawPolygon(bubble, rect, ctx){
		ctx.beginPath();
		_.each(bubble.points, function(point, index){
			var x = rect.x + point.x * rect.width;
			var y = rect.y + point.y * rect.height;
			if(index === 0){
				ctx.moveTo(x, y);
			}else{
				ctx.lineTo(x, y);
			}
		});
		ctx.closePath();
	}

	function drawRectangleWithArrows(ctx, frameWidth, frameHeight, topArrow, rightArrow, bottomArrow, leftArrow, arrowX, arrowY, arrowSize, arrowBaseSize, cornerRadius){

		//TOP
		ctx.moveTo(cornerRadius, 0);
		if(topArrow){
			ctx.lineTo(arrowX - arrowBaseSize / 2, 0);
			ctx.lineTo(arrowX, -arrowSize);
			ctx.lineTo(arrowX + arrowBaseSize / 2, 0);
		}
		ctx.lineTo(frameWidth - cornerRadius, 0);
		ctx.quadraticCurveTo(
			frameWidth, 0,
			frameWidth, cornerRadius
		);
		//RIGHT
		if(rightArrow){
			ctx.lineTo(frameWidth, arrowY - arrowBaseSize / 2);
			ctx.lineTo(frameWidth + arrowSize, arrowY);
			ctx.lineTo(frameWidth, arrowY + arrowBaseSize / 2);
		}
		ctx.lineTo(frameWidth, frameHeight - cornerRadius);
		ctx.quadraticCurveTo(
			frameWidth, frameHeight,
			frameWidth - cornerRadius, frameHeight
		);
		//BOTTOM
		if(bottomArrow){
			ctx.lineTo(arrowX + arrowBaseSize / 2, frameHeight);
			ctx.lineTo(arrowX, frameHeight + arrowSize);
			ctx.lineTo(arrowX - arrowBaseSize / 2, frameHeight);
		}
		ctx.lineTo(cornerRadius, frameHeight);
		ctx.quadraticCurveTo(
			0, frameHeight,
			0, frameHeight - cornerRadius
		);
		//LEFT
		if(leftArrow){
			ctx.lineTo(0, arrowY + arrowBaseSize / 2);
			ctx.lineTo(-arrowSize, arrowY);
			ctx.lineTo(0, arrowY - arrowBaseSize / 2);
		}
		ctx.lineTo(0, cornerRadius);
		ctx.quadraticCurveTo(
			0, 0,
			cornerRadius, 0
		);
	}

	//Draws a circle + border + some text inside
	function drawLabelCircle(ctx, text, rect, color, textHeight, drawLight){
		//measure text
		var textWidth = ctx.measureText(text).width;
		var cx = rect.x + rect.width / 2;
		var cy = rect.y + rect.height / 2;

		if(drawLight){
			ctx.globalAlpha = 0.3;
		}else{
			ctx.globalAlpha = 1;
		}
		ctx.strokeStyle = 'white';
		ctx.fillStyle = color;
		ctx.lineWidth = 2;
		ctx.font = textHeight + 'px Arial';

		//draw circle
		ctx.beginPath();
		ctx.arc(cx, cy, 4 + Math.max(textWidth, textHeight) / 2, 0, 2 * Math.PI, false);
		ctx.fill();
		ctx.stroke();

		//Draw text
		ctx.fillStyle = 'white';
		ctx.fillText(text, Math.round(cx), Math.round(cy));

		ctx.globalAlpha = 1;
	}

	return {
		clearFrame: clearFrame,
		drawBubble: drawBubble,
		drawPolygon: drawPolygon,
		drawBubbleBackground: drawBubbleBackground,
		drawRectangleWithArrows: drawRectangleWithArrows,
		drawLabelCircle: drawLabelCircle
	};
}]);

angular.module('fangaApp.model').service('fontList',
function(){

	//NOTE: font file names are normalized, see fontUtils.js
	return {
		ALL_FONTS: [{
			'fontFamily': 'comicneue',
			'label': 'Comic Neue',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'comicneueangular',
			'label': 'Comic Neue Angular',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'helsinki',
			'label': 'Helsinki',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikahand',
			'label': 'Komika Hand',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'komikaparch',
			'label': 'Komika Parch',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikatext',
			'label': 'Komika Text',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'komikatextkaps',
			'label': 'Komika Text Kaps',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'komikatexttight',
			'label': 'Komika Text Tight',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'komikatitle',
			'label': 'Komika Title',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikatitleaxis',
			'label': 'Komika Title Axis',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikatitlekaps',
			'label': 'Komika Title Kaps',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikatitlepaint',
			'label': 'Komika Title Paint',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'komikatitlewide',
			'label': 'Komika Title Wide',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'sfcartoonisthand',
			'label': 'SF Cartoonist Hand',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'sfcartoonisthandsc',
			'label': 'SF Cartoonist Hand SC',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'normal'
				},
				{
					'fontWeight':'bold',
					'fontStyle':'italic'
				},
				{
					'fontWeight':'normal',
					'fontStyle':'italic'
				}
			]
		},
		{
			'fontFamily': 'universalfruitcake',
			'label': 'Universal Fruit Cake',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		},
		{
			'fontFamily': 'vtclettererpro',
			'label': 'VTC Letter Pro',
			'rules': [
				{
					'fontWeight':'normal',
					'fontStyle':'normal'
				}
			]
		}
	]};
});
angular.module('fangaApp.misc').service('fontUtils',
function(){

	function createFontRulesText(font, fontFolderUrl){
		var ruleStrings = [];
		if(font.rules){
			var fontFolder = fontFolderUrl+ '/' + font.fontFamily + '/';
			for(var i = 0; i < font.rules.length; i++){
				var rule = font.rules[i];
				var path = fontFolder;
				if(rule.fontWeight == 'normal'){
					if(rule.fontStyle == 'normal'){
						path += 'regular';
					}else{
						path += 'italic';
					}
				}else{
					if(rule.fontStyle == 'normal'){
						path += 'bold';
					}else{
						path += 'bolditalic';
					}
				}
				path += '/font';

				// Rule style based on http://www.fontsquirrel.com/ syntax.
				ruleStrings.push(
					'@font-face {',
					' font-family:' + font.fontFamily + ';',
					' font-weight:' + rule.fontWeight + ';',
					' font-style:' + rule.fontStyle + ';',
					' src: url("' + path + '.eot ");',
					' src: url("' + path + '.eot?#iefix") format("embedded-opentype"),',
					'  url("' + path + '.woff") format("woff"),',
					'  url("' + path + '.ttf") format("truetype"),',
					'  url("' + path + '.svg#sf_cartoonist_handbold") format("svg");',
					'}'
				);
			}
		}
		return ruleStrings.join('\n');
	}

	function createFontUsageHTML(font){
		var html = '';
		if(font.rules){
			for(var i = 0 ; i < font.rules.length ; i++){
				var rule = font.rules[i];
				html += '<span style="' +
					'font-family:' + font.fontFamily + '; ' +
					'font-weight:' + rule.fontWeight + '; ' +
					'font-style:' + rule.fontStyle + '; ' +
					'">T</span>';
			}
		}
		return html;
	}

	return {
		initializeFontRules: function(fonts, createUsageElements, fontFolderUrl){
			fontFolderUrl = fontFolderUrl || '';
			
			//Create (or recycle) the style tag
			var STYLE_ELEMENT_ID = 'fanga-font-rules-style-element';
			var styleElement = document.getElementById(STYLE_ELEMENT_ID);
			if(!styleElement){
				styleElement = document.createElement('style');
				styleElement.id = STYLE_ELEMENT_ID;
				styleElement.type = 'text/css';
				document.getElementsByTagName('head')[0].appendChild(styleElement);
			}
			var rules = _.map(fonts, function(font){
				return createFontRulesText(font, fontFolderUrl);
			});
			styleElement.innerHTML = rules.join('\n');

			if(createUsageElements){
				//We create text elements that use the fonts in order to make the browser load them ASAP
				
				//Create (or recycle) some font usage div to force the font files downloading
				var FONT_USAGE_DIV_ID = 'fanga-font-usage-div';
				var fontUsageElement = document.getElementById(FONT_USAGE_DIV_ID);
				if(!fontUsageElement){
					fontUsageElement = document.createElement('div');
					fontUsageElement.id = FONT_USAGE_DIV_ID;
					fontUsageElement.style.visibility = 'hidden';
					document.body.appendChild(fontUsageElement);
				}
	
				var usageHtmls = _.map(fonts, function(font){
					return createFontUsageHTML(font);
				});
				fontUsageElement.innerHTML = usageHtmls.join('\n');
			}
		}
	};
});

angular.module('readerTools').service('fontManager', ["fontList", "fontUtils", function(fontList, fontUtils){
	var FAILURE_DELAY_MS = 3000;
	
	function loadFonts(fontFamilies, fontFolderUrl, callback){
		//font objects used to create the CSS rules
		var fontAssets = [];
		//font objects used by the FontLoader lib
		var fontLoaderFonts = [];
		_.each(fontFamilies, function(fontFamily){
			var fontAsset = _.findWhere(fontList.ALL_FONTS, {fontFamily:fontFamily});
			if(fontAsset){
				fontAssets.push(fontAsset);
				
				// FontLoader lib expects its fonts to have this structure: {family:'komikahand',weight:'normal',style:'normal'}
				_.each(fontAsset.rules, function(fontAssetRule){
					fontLoaderFonts.push({
						family : fontFamily,
						weight: fontAssetRule.fontWeight,
						style: fontAssetRule.fontStyle
					});
				});
			}
		});
		
		//create the CSS font rules that will reference those fonts
		fontUtils.initializeFontRules(fontAssets, false, fontFolderUrl);
		
		//make FontLoader check font loading
		var fontLoader = new FontLoader(fontLoaderFonts, {
			"fontLoaded": function(font) {
				// One of the fonts was loaded
				console.log("[fontManager] font loaded: " + font.family);
			},
			"complete": function(error) {
				callback(error);
			}
		}, FAILURE_DELAY_MS);
		fontLoader.loadFonts();
	}
	
	return {
		loadFonts: loadFonts
	};
}]);
(function() {
	var lastTime = 0;
	var vendors = ['webkit', 'moz'];
	for(var x = 0; x < vendors.length && !window.requestAnimationFrame; ++x) {
		window.requestAnimationFrame = window[vendors[x]+'RequestAnimationFrame'];
		window.cancelAnimationFrame =
			window[vendors[x]+'CancelAnimationFrame'] || window[vendors[x]+'CancelRequestAnimationFrame'];
	}

	if (!window.requestAnimationFrame)
		window.requestAnimationFrame = function(callback, element) {
			var currTime = new Date().getTime();
			var timeToCall = Math.max(0, 16 - (currTime - lastTime));
			var id = window.setTimeout(function() { callback(currTime + timeToCall); },
				timeToCall);
			lastTime = currTime + timeToCall;
			return id;
		};

	if (!window.cancelAnimationFrame)
		window.cancelAnimationFrame = function(id) {
			clearTimeout(id);
		};
}());
angular.module('readerTools').service('spreadDisplayer', ["fontManager", "drawUtils", function(fontManager, drawUtils){
	var fullSpread = null;
	var currentCallBack = null;
	var currentSpread = null;
	
	function init(){
		//we use the player spread class to do the display
		fullSpread = new fangaPreview.FullSpread(
			//for the moment, we assume image path is already absolute
			function(image){
				return image.path;
			},
			drawUtils
		);
		fullSpread.spreadDisplayed.add(onSpreadDisplayed);
		document.body.appendChild(fullSpread.element);
	}
	
	function display(spread, callback){
		currentSpread = spread;
		currentCallBack = callback;

        var animListener = null;
        var animListenerCallback = function(){
               var canvas = document.getElementById("box");
               canvas.removeEventListener("animationend", animListener, false);
               canvas.removeEventListener("webkitAnimationEnd", animListener, false);
               currentCallBack();
           };

        animListener = function AnimationListener(){
                setTimeout(animListenerCallback, 0);
            };

        var canvas = document.getElementById("box");
        canvas.style.webkitAnimation = 'none';
        canvas.addEventListener("animationend", animListener, false);
        canvas.addEventListener("webkitAnimationEnd", animListener, false);

		fullSpread.setSpread(spread);
		fullSpread.setIsVisibleInScreen(true);
	}
	
	function onSpreadDisplayed(spread){
		if(spread == currentSpread){
            var anim = document.getElementById("box");
            anim.style.webkitAnimation = 'none';
            setTimeout(function() {
                anim.style.webkitAnimation = '';
            }, 10);
//			currentCallBack();
		}
	}

	init();
	
	return {
		display: display
	};
	
}]);